/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	BlurXY
//
// COMMENTS: Use a Rayz IM to blur an image separately in X & Y
//
*/

#include <stdio.h>
#include <CPI.h>
#ifdef WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include "BlurXY.h"

// menu of actions for optional second input
static const char *actionmenu[] =
{
	"Mix",
	"On/Off",
	NULL
};

// menu of channels to use from optional second input
static const char *chanmenu[] =
{
	"Red channel",
	"Green channel",
	"Blue channel",
	"Alpha channel",
	NULL
};


////////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
BlurXYInit( void )
{
	cpiAddFloatPair( "blur_size", "Blur Size",
					 "x", "X", 0.01,
					 "y", "Y", 0.01,
					 0.0, 1.0,
					 CPI_PARM_ANIMATABLE | CPI_PARM_FULL_SIZE_XY,
					 NULL, NULL );
	cpiAddChannelSelect( NULL, NULL );
	cpiAddMenu( "mask_action", "Mask Action", 0, actionmenu, NULL, NULL );
	cpiAddMenu( "mask_channel", "Mask Channel", 0, chanmenu, NULL, NULL );

	// set up the node icon
	cpiRegisterIconData( "BlurXY.tif", BlurXY_SIZEX, BlurXY_SIZEY,
						 BlurXY_data, 255, 0, 255 );

	// define the name of the node icon
	cpiDefineIcon( "Nodes/BlurXY", "BlurXY.tif",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );
	
	return CPI_TRUE;
}


/////////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
BlurXYShutDown( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////////
// This routine gets called whenever the input changes. When
// that happens, we have to see if there is (still) a 2nd input
// and then enable/disable accordingly.

static void
BlurXYInputChanged( CPI_PrivateData	 /*handle*/,
					CPI_Uint32 		 /*inputnum*/ )
{
	CPI_Uint32			numInputs	= 0;
	CPI_Bool			enable		= CPI_FALSE;

	cpiGetNumInputs( &numInputs );

	// enable/disable the mask parameters depending on the presence
	// if the mask input
	if ( numInputs == 2 )
		enable = CPI_TRUE;

	cpiEnableParameter( "mask_action", enable );
	cpiEnableParameter( "mask_channel", enable );
}


//////////////////////////////////////////
// This optional routine defines help labels for the inputs.
// Also, the first letter of each label is used to label the
// input connectors.

static const char *
BlurXYInputLabel( CPI_PrivateData	 /*handle*/,
				  CPI_Uint32   		 input )
{
	if ( input == 0 )
		return( "Input image" );
	else
		return( "Mask image (optional)" );
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
BlurXYExec( CPI_PrivateData	 /*handle*/,
			CPI_Float32 	 time,
			CPI_Uint8		 quality,
			CPI_Uint32		 /*output*/,
			CPI_Bool		 /*viewerOutput*/,
			CPI_Float32		 scaleX,
			CPI_Float32		 scaleY )
{
	CPI_ImageOp		 	 myInputs[2];
	CPI_Uint32		 	 numInputs;
	CPI_Int32		 	 action;
	CPI_Float32			 ksizex, ksizey;
	CPI_ImageContext	 inputContext;
	CPI_ImageOp		 	 retval 	= NULL;
	CPI_PrivateData	 	 opData		= NULL;
	blurParms			*blurData	= NULL;
	
	// find out how many inputs we have connected
	cpiGetNumInputs( &numInputs );

	// get the first one, and then if there is a second,
	// get that too
	myInputs[0] = cpiGetInputOp( 0, time, quality, scaleX, scaleY  );
	myInputs[1] = NULL;
	if ( numInputs > 1 )
		myInputs[1] = cpiGetInputOp( 1, time, quality, scaleX, scaleY );

	if ( !cpiGetInputContext( &inputContext, 0, time ) )
	{
		cpiError( "Couldn't get information about required input" );
		return NULL;
	}
	
	// Since we are using a RAYZ internal IM, we didn't have to define
	// a structure to hold the parameters. One is provided for us when
	// we issue the following call.
	opData 		= cpiCreatePrivateData( "blur" );
	blurData	= (blurParms *)opData;

	if ( NULL != blurData )
	{
		// unload params from user interface and put them into the
		// parameter structure. The documentation contains a list
		// of all builtin RAYZ IMs and their internal parameter names.
		if ( cpiGetFloat( &ksizex, "blur_size.x", time ) &&
			 cpiGetFloat( &ksizey, "blur_size.y", time ) &&
			 cpiGetChannelSelect( &blurData->channelSelect ) &&
			 cpiGetInteger( &action, "mask_action", time ) &&
			 cpiGetInteger( &blurData->maskChannel, "mask_channel", time ) )
		{
			// add the 'blur' op to the list
			// in this case, we are using the Rayz internal 'blur' op
			// so this operation is not defined in this file.
			// The documentation gives a list of Rayz ops and their
			// private data which you may use.
			blurData->maskType	= (CPI_MaskType)action;
			blurData->size[0]	= inputContext.sizeX;
			blurData->size[1]	= inputContext.sizeY;
			blurData->offset[0]	= inputContext.offsetX;
			blurData->offset[1]	= inputContext.offsetY;

			// set the kernel size in pixels, relative to the current
			// image size
			blurData->kernelSize[0] = ksizex * inputContext.sizeX;
			blurData->kernelSize[1] = ksizey * inputContext.sizeY;
			
			if ( NULL != myInputs[0] )
				retval = cpiAddImageOp( "blur", blurData, myInputs, 2 );
			else
			{
				cpiError( "Input A is not connected" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
		{
			cpiError( "Couldn't get BlurXY parameter(s)" );
			cpiDestroyPrivateData( opData );
		}
	}
	else
		cpiError( "Couldn't get BlurXY private data" );
		
	return retval;
}


//////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"blurxy",		// node name
		"BlurXY",		// menu name
		"Silicon Grail",
		"2.2", 
		NULL, 
		NULL
	},

	BlurXYInit,		// NodeInit
	BlurXYShutDown,	// NodeShutDown
	NULL,			// CreateInstance
	NULL,			// DestroyInstance
	NULL,			// ParmChanged
	NULL,			// InputAdded
	BlurXYInputChanged,
	NULL,			// InputRemoved
	NULL,			// GetRegion
	NULL,			// GetFullSize
	NULL,			// GetRange
	NULL,			// GetErrors
	BlurXYExec,		// Node Execute
	BlurXYInputLabel,
	NULL,			// OutputLabel
	NULL,			// ViewerLabel
	NULL,			// ViewerOutputs
	1,				// min inputs
	2,				// max inputs
	1,				// min outputs
	1,				// max outputs
	CPI_FALSE		// is the node growable? (like Mcomp)
};


///////////////////////////////////////
// This is called once on startup. If you were going to read
// in an outside table, you would do it here

static CPI_Bool
BlurXYScriptInit( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////
// This cleans up on exit

static CPI_Bool
BlurXYScriptShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////
// Output image out node desc to the specified device

static CPI_Bool
BlurXYScriptProcess( CPI_DeviceHandle	  device,
			         const char			 *inputName,
			         CPI_Uint32			  numArgs,
			         const char * const	 *arguments,
			         char				**outputName )
{
	char		*width		= NULL;
	char		*height		= NULL;
	bool		 widthFr	= false;
	bool		 heightFr	= false;
	bool		 isOkay		= true;

	if ( 1 > numArgs || NULL == arguments[0] )
	{
		cpiError( "BlurXY command missing width." );
		isOkay	= false;
	}
	else
	{
		widthFr = cpiCheckSizeForFraction( arguments[0], &width );
		isOkay	= NULL != width;
	}

	if ( 2 > numArgs || NULL == arguments[1] )
	{
		cpiError( "BlurXY command missing height." );
		isOkay	= false;
	}
	else
	{
		heightFr	= cpiCheckSizeForFraction( arguments[1], &height );
		isOkay		= NULL != height;
	}

	if ( 2 < numArgs )
	{
		cpiError( "Too many arguments for blurxy command." );
		isOkay	= false;
	}

	if ( isOkay )
	{
		ostrstream	 buf;
		char		*nodeName	= cpiGetUniqueNodeName();

		// Add blurxy node
		buf	<< nodeName << " = blurxy( " <<  inputName << " )\n"
			<< "{\n";
		if ( widthFr )
			buf << "  blur_size.x: " << width << ";\n";
		else
		{
			buf << "  blur_size.x: constant 0@-10000 expression( \""
				<< width << "/$W\" ) 0@10000 constant;\n";
		}

		if ( heightFr )
			buf << "  blur_size.y: " << height << ";\n";
		else
		{
			buf << "  blur_size.y: constant 0@-10000 expression( \""
				<< height << "/$H\" ) 0@10000 constant;\n";
		}
			
		buf << "}\n\n";

		if ( buf.pcount() == cpiWrite( device, buf.str(), buf.pcount() ) )
		{
			(*outputName)	= nodeName;
		}
		else
		{
			cpiError( "Error writing blurxy node to file." );
			cpiFree( nodeName );
			isOkay	= false;
		}
	}

	cpiFree( width );
	cpiFree( height );

	return ( isOkay ) ? CPI_TRUE : CPI_FALSE;
}


////////////////////////////////////
// As with nodes, scripts are described by a structure like
// this one. This must be present.

static RPI_ScriptInfo scriptinfo = {
	{
		"blurxy",                       // internal name
		"BlurXY"    ,                   // name which appears in the menu
		"Silicon Grail",                // author
		"2.2",                          // version
		"",                             // help URL
		""                              // icon name
	},
	BlurXYScriptInit,                   // init function
	BlurXYScriptShutDown,               // shutdown function
	BlurXYScriptProcess,                // process function
	"--blurxy <xsize>[f|p] <ysize>[f|p]",   // usage
	CPI_FALSE                           // render by default
};


/////////////////////////////////////////////////
// register the node
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval = cpiRegisterNode( &ninfo );

	if ( retval )
		retval = cpiRegisterScriptOption( &scriptinfo );
	
	return retval;
}


//////////////////////////////////////////////
// unregister the node

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterNode( "blurxy" );
}
