/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Callback
//
// COMMENTS: Demonstrate setting and using a callback button
//			 in the Rayz GUI
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _callbackState
{
	CPI_Int32	 direction;
	CPI_Float32	 size;
} callbackState;


//////////////////////////////////////
// This is the function to be called when the button is pressed
// It prints out its arguments, and shows how to programmatically
// set the values of other GUI widgets
// Notice that when setting widget values from the software (as opposed
// to having the user do it directly), you need to set the final arg
// to CPI_FALSE. This tells Rayz not to save this value to the .rayz
// file, because it is not user set.

static void
CB_CallBack( const char 		*nodename,
			 const char 		*parmname,
			 CPI_Float32	 	 time,
			 CPI_PrivateData	 data )
{
	printf ("Got a button callback for node '%s', parm '%s' at time %5.5f\n", 
			 nodename, parmname, time );
	cpiSetInteger( 2, "direction", time, CPI_FALSE );
	cpiSetFloat( time, "size", time, CPI_FALSE );
}


//////////////////////////////////////
// menu of directions

static const char *dirmenu[]= 
{
	"up",
	"down",
	"left",
	"right",
	NULL
};


////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
CB_Init( void )
{
	cpiAddFloat( "size", "Size", 
				 1.0, 
				 0.01, 
				 2, 
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				 "A float value to set",
				 NULL );
	cpiAddMenu( "direction", "Direction", 
				0, 
				dirmenu,
				"A menu to choose from",
				NULL );

	// this is the button definition
	cpiAddButton( "button", "Reset Values", 
				  CB_CallBack,
				  CPI_PARM_DEFAULT,
				  "Test callbacks by setting the direction to 'left'",
				  NULL );
	
	return CPI_TRUE;
}

///////////////////////////////////////////
// called to clean up any messes - in most cases it does nothing

static CPI_Bool
CB_ShutDown( void )
{
	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
CB_Exec( CPI_PrivateData	 /*handle*/,
		 CPI_Float32 		 myTime,
		 CPI_Uint8      	 quality,
		 CPI_Uint32     	 /*nodeOutput*/,
		 CPI_Bool       	 /*viewerOutput*/,
		 CPI_Float32    	 scaleX,
		 CPI_Float32    	 scaleY )
{
	CPI_PrivateData	 opData;
	CPI_ImageOp		 retval		= NULL;
	CPI_ImageOp 	 myInput 	= NULL;
	callbackState	*cbData		= NULL;

	myInput = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	// myInput now points to the first input
	// inputs are numbered from 0
	if ( myInput == NULL )
		return retval;

	// If we have a valide input, then allocate a parameter structure.
	// This is as defined in the earlier typedef
	opData	= cpiCreatePrivateData( "callback" );
	cbData	= (callbackState *)opData;

	if ( NULL != cbData )
	{
		// get the user parameter value and stuff it into the
		// structure. 
		if ( cpiGetInteger( &cbData->direction, "direction", myTime ) &&
			 cpiGetFloat( &cbData->size, "size", myTime ) )
		{
			// add this op to the exec list
			// the name of the op (callback) is defined in the
			// RPI_ImageOpInfo structure later in the file
			retval = cpiAddImageOp( "callback", cbData, &myInput, 1 );
		}
		else
		{
			cpiError( "Couldn't get Callback parameters(s)" );
			cpiDestroyPrivateData( opData );
		}
	}
	else
		cpiError( "Couldn't get Callback private data" );
	
	return retval;
}


////////////////////////////////////////////

static RPI_NodeInfo ninfo = {
	{
		"callback",		// node name
		"Callback",		// menu name
		"Joe Programmer",
		"2.2"
	},
	CB_Init,		// Init
	CB_ShutDown,	// ShutDown
	NULL,			// CreateInstance
	NULL,			// DestroyInstance
	NULL,			// ParmChanged
	NULL,			// InputAdded
	NULL,			// InputChanged
	NULL,			// InputRemoved
	NULL,			// GetRegion
    NULL,			// GetFullSize
	NULL,			// GetRange
	NULL,			// GetErrors
	CB_Exec,		// Execute
	NULL,			// InputLabel
	NULL,			// OutputLabel
	NULL,			// ViewerLabel
	NULL,			// ViewerOutputs
	1, 				// min inputs
	1,				// max inputs
	1,				// min outputs
	1,				// max outputs
	CPI_FALSE		// is growable
};


///////////////////////////////////////////////////////////
// Image Op Functions /////////////////////////////////////
///////////////////////////////////////////////////////////

////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

template <class T>
static void simple( CPI_Image  &input,
					CPI_Image  *result,
					T           /*max*/)
{
	CPI_Int32    x, y, endX, endY, i, j = 0;
	CPI_Int32    chans = result->myContext.channels;
	T           *pin, *pout;

	// this routine just copies its input to its output
	endX    = result->myContext.offsetX + result->myContext.sizeX;
	endY    = result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );
		j 		= 0;

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// copy each channel
			for ( i = 0; i < chans; i++, j++ )
				pout[j] = pin[j];
		}
	}
}


////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
CB_CreateInstance( void )
{
	callbackState	*retval = (callbackState *)cpiAlloc( sizeof( callbackState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.


static CPI_Bool
CB_DestroyInstance( CPI_PrivateData	handle )
{
	CPI_Bool	 retval = CPI_FALSE;

	if ( NULL != handle )
	{
		callbackState	*state = (callbackState *)handle;

		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

static CPI_Bool
CB_Process( CPI_PrivateData	 handle,
			CPI_Image		*result,
		    CPI_Image	 	 inputs[],
		    CPI_Uint32	 	 /*numInputs*/ )
{
	CPI_Int32	 	 direction;
	CPI_Float32	 	 size;
	CPI_Bool	 	 retval	= CPI_TRUE;
	callbackState	*state	= (callbackState *)handle;

	size 		= state->size;
	direction	= state->direction;
	
	if ( result->myContext.bitsPerPel == 8 )
		simple( inputs[0], result, (CPI_Uint8)255 );
	else if ( result->myContext.bitsPerPel == 16 )
		simple( inputs[0], result, (CPI_Uint16)65535 );
	else
		simple( inputs[0], result, (CPI_Float32)1.0F );

	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"callback",			// name of op to add to list
		"",					// not used
		"Silicon Grail",	// author
		"2.2",				// version
		NULL, NULL
	},
	CB_CreateInstance,
	CB_DestroyInstance,
	CB_Process,			// image execution process
	NULL,				// InputRegion
	NULL,				// output region
	NULL,				// control processing
	NULL,				// setup processing
	NULL,				// cleanup processing
	1,					// min inputs
	CPI_FALSE,			// multiproc?
	CPI_FALSE			// disjoint regions?
};


//////////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "callback" );
	cpiUnregisterNode( "callback" ); 
}
