/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	ChannelSplit
//
// COMMENTS: Take an RGBA image and split it into two outputs, RGB and A
//           This demonstrates a node with more than one output.
//
*/

#include <stdio.h>
#include <CPI.h>
#include "ChannelSplit.h"



////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _splitState
{
	CPI_Int32	 channels;
} splitState;


///////////////////////////////////////


static CPI_Bool
ChannelSplitInit( void )
{
	// set up the node icon
	cpiRegisterIconData( "ChannelSplit.tif", ChannelSplit_SIZEX, ChannelSplit_SIZEY,
	  					  ChannelSplit_data, 255, 0, 255 );
	// define the name of the checker icon
	cpiDefineIcon( "Nodes/Channel Split", "ChannelSplit.tif",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );

	return CPI_TRUE;
}


/////////////////////////////////////


static CPI_Bool
ChannelSplitShutDown( void )
{
	return CPI_TRUE;
}


/////////////////////////////////////


static CPI_Bool
ChannelSplitGetRegion( CPI_PrivateData	 /*handle*/,
					   CPI_ImageContext	*region,
					   CPI_Float32		 myTime,
					   CPI_Uint32		 output,
					   CPI_Bool	 	 	 /*viewerOutput*/ )
{
	CPI_Bool	retval = CPI_FALSE;

	if ( cpiGetInputContext( region, 0, myTime ) && 
		 4 == region->channels )
	{
		if ( output == 0 )
		{
			region->channels	= 3;
			retval 				= CPI_TRUE;
		}
		else if ( output == 1 )
		{
			region->channels 	= 1;
			retval				= CPI_TRUE;
		}
		else
			cpiError( "Channel Split got a request for an invalid output" );
	}
	else
		cpiError( "Channel Split couldn't get input or input didn't have 4 "
				  "channels" );

	return retval;
}


//////////////////////////////////////


static CPI_Bool
ChannelSplitGetErrors( CPI_PrivateData	 /*handle*/,
					   CPI_Float32 		 time )
{
	CPI_ImageContext 	 input;
	CPI_Bool			 retval	= CPI_FALSE;

	if ( cpiGetInputContext( &input, 0, time ) )
	{
		if ( 4 <= input.channels )
			retval = CPI_TRUE;
		else
			cpiError( "Input must have at least 4 channels!" );
	}
	else
		cpiError( "Couldn't get input context!" );

	return retval;
}


/////////////////////////////////////////


static CPI_ImageOp
ChannelSplitExec( CPI_PrivateData	 /*handle*/,
				  CPI_Float32		 time,
				  CPI_Uint8			 quality,
				  CPI_Uint32		 output,
				  CPI_Bool			 /*viewerOutput*/,
				  CPI_Float32	 	 scaleX,
				  CPI_Float32		 scaleY )
{
	CPI_ImageOp	retval	= NULL;
	CPI_ImageOp	inputOp	= cpiGetInputOp( 0, time, quality, scaleX, scaleY );

	if ( NULL != inputOp )
	{
		CPI_PrivateData	 opData;
		splitState		*opState	= NULL;

		opData	= cpiCreatePrivateData( "channelsplit" );
		opState	= (splitState *)opData;

		if ( NULL != opData )
		{
			if ( 0 == output )
				opState->channels = 3;
			else if ( 1 == output )
				opState->channels = 1;
			
			retval = cpiAddImageOp( "channelsplit", opState, &inputOp, 1 );
		}
		else
			cpiError( "Couldn't get ChannelSplit private data" );
	}

	return retval;
}


/////////////////////////////////////
// label the node inputs - this is optional
// Input is the number of the input you want to provide a
// label for, starting at 0 - this label shows up in the
// status line when the user mouses over the connector.
// The first letter of each string is used to label the input
// connector.


static const char *
ChannelSplitOutputLabel( CPI_PrivateData	 /*handle*/,
						 CPI_Uint32			 output )
{
	const char	*retval	= NULL;

	if ( 0 == output )
		retval = "RGB Output";
	else if ( 1 == output )
		retval = "Alpha Output";

	return retval;
}


/////////////////////////////////////


static RPI_NodeInfo ninfo = {
	{
		"channelsplit",
		"Channel Split",
		"Silicon Grail",
		"2.2"
	},
	ChannelSplitInit,
	ChannelSplitShutDown,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	ChannelSplitGetRegion,
	NULL,
	NULL,
	ChannelSplitGetErrors,
	ChannelSplitExec,
	NULL,
	ChannelSplitOutputLabel,
	NULL,
	NULL,
	1,
	1,
	2,
	2,
	CPI_FALSE
};


///////////////////////////////////////////////////////////////////////////
// Image Op Functions /////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////
//
// These are the actual image manipulation routines. They are templated
// functions so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.
//
// You are guaranteed that the two images (and the result) will
// all be the same size if the 'disjoint regions' flag is CPI_FALSE
// in the RPI_ImageOpInfo structure below. See later note.


template <class T> static
void split_alpha( CPI_Image	*result,
				  CPI_Image	*input,
				  T			 /*max*/ )
{
	CPI_Uint32	 endX, endY;
	T			*out	= (T *)result->data;
	T			*in		= (T *)input->data;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;

	for ( CPI_Uint32 y = result->myContext.offsetY; y < endY; ++y )
	{
		out = (T *)cpiGetLine( result, y );
		in	= (T *)cpiGetLine( input, y );
		
		for ( CPI_Uint32 x = result->myContext.offsetX; x < endX; ++x )
		{

			in += 3;
			*out++ = *in++;
		}
	}
}


//////////////////////////////////////////


template <class T> static
void split_rgb( CPI_Image	*result,
				CPI_Image	*input,
				T			 /*max*/ )
{
	CPI_Uint32	 endX, endY;
	T			*out	= (T *)result->data;
	T			*in		= (T *)input->data;
	CPI_Uint32	 skip	= ( input->myContext.channels - 3 );

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;

	for ( CPI_Uint32 y = result->myContext.offsetY; y < endY; ++y )
	{
		out = (T *)cpiGetLine( result, y );
		in	= (T *)cpiGetLine( input, y );
		
		for ( CPI_Uint32 x = result->myContext.offsetX; x < endX; ++x )
		{
			*out++ = *in++;
			*out++ = *in++;
			*out++ = *in++;
			in += skip ;
		}
	}
}


/////////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
ChannelSplitCreateInstance( void )
{
	splitState	*retval = (splitState *)cpiAlloc( sizeof( splitState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.


static CPI_Bool
ChannelSplitDestroyInstance( CPI_PrivateData handle )
{
	splitState *state = (splitState *)handle;

	if ( NULL != state )
		cpiFree( state );

	return CPI_TRUE;
}


////////////////////////////////////////


static CPI_Bool
ChannelSplitOutputRegion( CPI_PrivateData	 handle,
						  CPI_ImageContext	*output,
						  CPI_ImageContext	 inputs[],
						  CPI_Uint32		 numInputs )
{
	CPI_Bool	 retval = CPI_FALSE;
	splitState	*state	= (splitState *)handle;

	if ( numInputs > 0 && NULL != state )
	{ 
		*output = inputs[0];

		output->channels = state->channels;
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth


static CPI_Bool
ChannelSplitProcess( CPI_PrivateData	 handle,
					 CPI_Image			*result,
					 CPI_Image	 	 	 inputs[],
					 CPI_Uint32	 	 	 numInputs )
{
	CPI_Bool	 retval	= CPI_FALSE;
	splitState	*state	= (splitState *)handle;
	
	if ( NULL != result && numInputs == 1 )
	{
		CPI_ImageContext	c = inputs[0].myContext;

		retval	= CPI_TRUE;
		
		if ( retval )
		{
			// Call the appropriate routine, depending on image types
			if ( state->channels == 1 )
			{
				if ( c.bitsPerPel == 8 )
					split_alpha( result, &inputs[0], (CPI_Uint8)255 );
				else if ( c.bitsPerPel == 16 )
					split_alpha( result, &inputs[0], (CPI_Uint16)65535 );
				else if ( c.bitsPerPel == 32 )
					split_alpha( result, &inputs[0], (CPI_Float32)1.0 );
				else
				{
					cpiError( "Inputs image has an invalid bit depth!" );
					retval = CPI_FALSE;
				}
			}
			else if ( state->channels == 3 )
			{
				if ( c.bitsPerPel == 8 )
					split_rgb( result, &inputs[0], (CPI_Uint8)255 );
				else if ( c.bitsPerPel == 16 )
					split_rgb( result, &inputs[0], (CPI_Uint16)65535 );
				else if ( c.bitsPerPel == 32 )
					split_rgb( result, &inputs[0], (CPI_Float32)1.0 );
				else
				{
					cpiError( "Inputs image has an invalid bit depth!" );
					retval = CPI_FALSE;
				}	
			}
			else
			{
				cpiError( "Invalid channels value!" );
				retval = CPI_FALSE;
			}
		}

	}

	return retval;
}


///////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.
// Required entries are marked, the rest are optional. Fill in
// NULL if there is no entry for an optional function.


static RPI_ImageOpInfo opinfo =
{
	{					// this structure is required
		"channelsplit",
		"Channel Split",
		"Silicon Grail",
		"2.2"
	},
	ChannelSplitCreateInstance,			// required
	ChannelSplitDestroyInstance,
	ChannelSplitProcess,				// required
	NULL,								// Input Region
	ChannelSplitOutputRegion,
	NULL,								// Control Processing
	NULL,								// Setup Processing
	NULL,								// Cleanup Processing
	1,									// min number inputs required
	CPI_FALSE,							// Can this plugin multiproc?
	CPI_FALSE							// Disjoint regions?
};
	

//////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool	retval = CPI_TRUE;

	retval = cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;
}


///////////////////////////////////////
// unregister node and image op


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "channelsplit" );
	cpiUnregisterNode( "channelsplit" );
}
