/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Checker
//
// COMMENTS: This illustrates a source node (one with no inputs)
//
*/

#include <stdio.h>
#include <string.h>
#include <CPI.h>

#include "Checker.h"


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _checkerState
{
	CPI_Int32	 size[2];
	CPI_Float32	 scale[2];
	CPI_Int32	 channels;
	CPI_Int32	 bitdepth;
	CPI_Int32	 squareSize[2];
	CPI_Float32	 bgColor[4];
	CPI_Float32	 fgColor[4];	
} checkerState;


//////////////////////////////////////
// menu of which channels to create

static const char *channelmenu[]= 
{
	"RGB Alpha",
	"RGB",
	"Alpha only",
	NULL
};


///////////////////////////////////////
// menu of which bit depth to create

static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////
// default color arrays

static CPI_Float32 fgsqs[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 bgsqs[] = { 0.0F, 0.0F, 0.0F, 0.0F };


/////////////////////////////////////////////////////
// Node Functions ///////////////////////////////////
/////////////////////////////////////////////////////

// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
CheckerInit( void )
{
	cpiAddColorGroup( "sq", "Square Col", fgsqs, 4, CPI_PARM_DEFAULT, 
						"Color of squares", NULL );
	cpiAddColorGroup( "bg", "Background", bgsqs, 4, CPI_PARM_DEFAULT, 
						"Color between squares", NULL );
	cpiAddIntPair( "sqsize", "Square Size",
				   "x", "X", 64,
				   "y", "Y", 64,
				   1, 4096,
				   CPI_PARM_DEFAULT,
				   "Size of squares", NULL );

	cpiAddIntPair( "size", "Size",
				   "w", "Width", 720,
				   "h", "Height", 576,
				   1, 4096,
				   CPI_PARM_CLAMPMIN,
				   "Size of output image", NULL );
	cpiAddMenu( "bits", "Bit Depth", 0, bitdepthmenu, NULL, NULL );
	cpiAddMenu( "chans", "Channels", 0, channelmenu, NULL, NULL );
	cpiAddSequenceRange( 1, 1,
						CPI_PARM_DEFAULT,
						"Range of frames to output", NULL );

	// set up the node icon
	 cpiRegisterIconData( "Checker.tiff", Checker_SIZEX, Checker_SIZEY,
	  					   Checker_data, 255, 0, 255 );
	// define the name of the checker icon
	cpiDefineIcon( "Nodes/Checker", "Checker.tiff",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );

	return CPI_TRUE;
}

///////////////////////////////////////////
// shutdown does nothing in this case

static CPI_Bool
CheckerShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////////
// CreateInstance is called each time a node of this type is
// created. We can use this function to discover the name of
// the current project. Then we can use that to get the
// value of a Project preference

#define XFULLSIZE_PATH  "/Settings/Default Full Size X"
#define YFULLSIZE_PATH  "/Settings/Default Full Size Y"

static CPI_PrivateData
CheckerInstance( void )
{
	const char	*projname;
	char		 fullPath[64];
	CPI_Int32	 xsize, ysize;

	// get the name of the current project
	if ( NULL == ( projname = cpiGetProjectName() ) )
	{	
		cpiError("Couldn't get Project Name" );
		return NULL;
	}

	// now create the paths to the preferences we want
	strcpy( fullPath, projname );
	strcat( fullPath, XFULLSIZE_PATH );
	if ( !cpiGetPreferenceInt( &xsize, fullPath ) )
	{
		cpiError( "Couldn't get X size" );
		return NULL;
	}

	strcpy( fullPath, projname );
	strcat( fullPath, YFULLSIZE_PATH );
	if ( !cpiGetPreferenceInt( &ysize, fullPath ) )
	{
		cpiError( "Couldn't get Y size" );
		return NULL;
	}

	// finally, set these sizes into the GUI widgets as the new
	// defaults for this instance of this node 
	if ( !cpiSetInteger( xsize, "size.w", 1.0F, CPI_FALSE ) ||
		 !cpiSetInteger( ysize, "size.h", 1.0F, CPI_FALSE ) )
	{
		cpiError( "Unable to set default size" );
		return NULL;
	}
		 	
	return NULL;
}


///////////////////////////////////////////
// GetRegion needs to be defined for nodes that generate their
// own images (as opposed to being given images to modify)
// This defines for the rest of Rayz the output 'context',
// that is, size, bit depth, etc.

static CPI_Bool
CheckerGetRegion( CPI_PrivateData	 /*handle*/,
				  CPI_ImageContext	*output,
				  CPI_Float32		 myTime,
				  CPI_Uint32		 /*nodeOutput*/,
				  CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		CPI_Int32	 bitsPerPel, channels;

		// get size from GUI and store in 'output'
		retval = ( cpiGetInteger( &output->sizeX, "size.w", myTime ) &&
				   cpiGetInteger( &output->sizeY, "size.h", myTime ) &&
				   cpiGetInteger( &bitsPerPel, "bits", myTime ) &&
				   cpiGetInteger( &channels, "chans", myTime ) );

		if ( retval )
		{
			output->offsetX = 0;
			output->offsetY = 0;

			switch ( bitsPerPel )
			{
				case 0:
					output->bitsPerPel = 8;
					break;
				case 1:
					output->bitsPerPel = 16;
					break;
				case 2:
					output->bitsPerPel = 32;	// ie, float
					break;
				default:
					cpiError( "Invalid bitsPerPel in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			// get user menu choice, and set value accordingly
			switch ( channels )
			{
				case 0:
					output->channels = 4;
					break;
				case 1:
					output->channels = 3;
					break;
				case 2:
					output->channels = 1;
					break;
				default:
					cpiError( "Invalid channels in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}
		}
		else
		{
			cpiError( "Couldn't get parameter!" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Frame Out of Range!" );
		retval = CPI_FALSE;
	}
	
	return retval;
}



////////////////////////////////////////


static void
CheckerParmChanged( CPI_PrivateData	 /*handle*/,
					const char 		*parmname )
{
	// If the parm name is NULL, that means the a new node has just been added
	// we want to set the seuqence parm to match the global frame range.
	if ( parmname == NULL )
	{
		CPI_Int32	 startFrame, endFrame;

		if ( cpiGetGlobalFrameRange( &startFrame, &endFrame ) )
			cpiSetSequenceParmRange( startFrame, endFrame );
	}
}


///////////////////////////////////////////
// Rayz may need to ask the plugin what size
// image it intends to create - define that here
// Nodes that create their own output need to
// define this routine

static CPI_Bool
CheckerGetFullSize( CPI_PrivateData	  /*handle*/,
					CPI_ImageContext *fullsize,
					CPI_Float32		  myTime,
					CPI_Uint32		  /*nodeOutput*/,
					CPI_Bool       	  /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool 	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( !cpiGetInteger( &fullsize->sizeX, "size.w", myTime ) ||
			 !cpiGetInteger( &fullsize->sizeY, "size.h", myTime ) )
		{
			cpiError( "Couldn't get size parameter" );
			retval = CPI_FALSE;
		}			
	}
	else
	{
		cpiError( "Frame Out Of Range" );
		retval = CPI_FALSE;
	}		
		
	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the parameters from the user interface
// 2. allocate metadata and copy the parameters to it
// 3. add the image op(s) to the list of operations

static CPI_ImageOp
CheckerExec( CPI_PrivateData	 /*handle*/,
			 CPI_Float32 	 	 myTime,
			 CPI_Uint8	 	 	 /*quality*/,
			 CPI_Uint32	 	 	 /*output*/,
			 CPI_Bool 	 	 	 /*viewerOutput*/,
			 CPI_Float32 	 	 scaleX,
			 CPI_Float32	 	 scaleY )
{
	CPI_Int32	startFrame, endFrame;
	CPI_Int32	channels, bitdepth;
	CPI_Int32	xsquare, ysquare;
	CPI_ImageOp	retval	= NULL;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		// These 4 lines define and allocate a block of data as
		// defined in the earlier typedef
		CPI_PrivateData	 opData;
		checkerState	*opState	= NULL;

		opData	= cpiCreatePrivateData( "checker" );
		opState	= (checkerState *)opData;
		
		if ( NULL != opState )
		{
			// get values from user and stuff into param structure
			if ( cpiGetInteger( &opState->size[0], "size.w", myTime ) &&
				 cpiGetInteger( &opState->size[1], "size.h", myTime ) &&
				 cpiGetInteger( &channels, "chans", myTime ) &&
				 cpiGetInteger( &bitdepth, "bits", myTime ) &&
				 cpiGetInteger( &xsquare, "sqsize.x", myTime ) &&
				 cpiGetInteger( &ysquare, "sqsize.y", myTime ) &&
				 cpiGetFloat( &opState->bgColor[0], "bg.red", myTime ) &&
				 cpiGetFloat( &opState->bgColor[1], "bg.green", myTime ) &&
				 cpiGetFloat( &opState->bgColor[2], "bg.blue", myTime ) &&
				 cpiGetFloat( &opState->bgColor[3], "bg.alpha", myTime ) &&
				 cpiGetFloat( &opState->fgColor[0], "sq.red", myTime ) &&
				 cpiGetFloat( &opState->fgColor[1], "sq.green", myTime ) &&
				 cpiGetFloat( &opState->fgColor[2], "sq.blue", myTime ) &&
				 cpiGetFloat( &opState->fgColor[3], "sq.alpha", myTime ) )
			{
				opState->scale[0]	= scaleX;
				opState->scale[1]	= scaleY;
				
				switch ( channels )
				{
					case 0:
						opState->channels	= 4;
						break;
					case 1:
						opState->channels	= 3;
						break;
					case 2:
						opState->channels	= 1;
						break;
				}

				switch ( bitdepth )
				{
					case 0:
						opState->bitdepth	= 8;
						break;
					case 1:
						opState->bitdepth	= 16;
						break;
					case 2:
						opState->bitdepth	= 32;
						break;
				}
				
				// Resize relevant parameters based on scale.
				// This happens when the user requests a "medium" or "low"
				// image, and when your node shows up in the Clip Editor -
				// the little thumbnails need to be scaled accordingly
				// Your node needs to
				// correctly output something of whatever requested scale.
				xsquare = cpiRound( xsquare * scaleX );
				ysquare = cpiRound( ysquare * scaleY );

				if (xsquare < 1)
					xsquare = 1;
				if (ysquare < 1)
					ysquare = 1;

				opState->squareSize[0]	= xsquare;
				opState->squareSize[1]	= ysquare;

				retval = cpiAddImageOp( "checker", opState, NULL, 0 );
			}
			else
			{
				cpiError( "Couldn't get parameters" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
			cpiError( "Couldn't get Checker private data" );
	}
	else
		cpiError( "Frame Out Of Range" );
		
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not defined are set to NULL. Routines
// which are required to be provided are marked as such.

static RPI_NodeInfo ninfo = {
	{						// this structure is required
		"checker",			// node name
		"Checker",			// menu name
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	CheckerInit,		// NodeInit		(required)
	CheckerShutDown,	// NodeShutDown
	CheckerInstance,	// CreateInstance
	NULL,				// DestroyInstance
	CheckerParmChanged, // ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	CheckerGetRegion,	// GetRegion
    CheckerGetFullSize,	// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	CheckerExec,		// Node Execute	(required)
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	0, 					// min input
	0,					// max input
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable	(eg Mcomp)
};


////////////////////////////////////////////////////
// Image Op Functions //////////////////////////////
////////////////////////////////////////////////////


////////////////////////////////////////
// This templated function does the actual work. By making
// it a template, it is only defined once, and the compiler
// creates a version for each of 8 bit, 16 and float inputs.

template <class T> static
void make_ramp( CPI_Image	*image,
				CPI_Int32	 fullY,
				CPI_Int32	 squares[], 
				CPI_Float32	 lvals[], 
				CPI_Float32	 dvals[],
				T 			 max )
{
	CPI_Int32	 x, y, chans = image->myContext.channels;
	CPI_Uint32	 a, b, c;
	CPI_Int32	 startX, startY, endX, endY;
	T			*ptr;
	struct {
		T	r, g, b, a;
	} sq[2];

	// init color table
	sq[0].r = (T) (lvals[0] * max);
	sq[0].g = (T) (lvals[1] * max);
	sq[0].b = (T) (lvals[2] * max);
	sq[0].a = (T) (lvals[3] * max);
	sq[1].r = (T) (dvals[0] * max);
	sq[1].g = (T) (dvals[1] * max);
	sq[1].b = (T) (dvals[2] * max);
	sq[1].a = (T) (dvals[3] * max);

	startX	= image->myContext.offsetX;
	startY	= image->myContext.offsetY;
	endX	= startX + image->myContext.sizeX;
	endY	= startY + image->myContext.sizeY;
	fullY  -= 1;		// this is used to invert Y coords
	
	for ( y = startY; y < endY; y++ )
	{
		// you should use this function to address scanlines
		// as it takes into account offsets and scale
		ptr = (T *)cpiGetLine( image, y );
		a = ( ( fullY - y ) / squares[1] ) % 2;	// which row
		for ( x = startX; x < endX; x++ )
		{
			b = ( x / squares[0] ) % 2;			// which column
			c = ( a + b ) % 2;
			// put in color, update pointer
			if ( chans >= 3 )
			{
				*ptr++ = sq[c].r;
				*ptr++ = sq[c].g;
				*ptr++ = sq[c].b;
			}
			// put in alpha if required
			if ( chans == 1 || chans == 4 )
				*ptr++ = sq[c].a;
		}
	}
}


////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
CheckerCreateInstance( void )
{
	checkerState *retval = (checkerState *)cpiAlloc( sizeof( checkerState ) );
	
	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.


static CPI_Bool
CheckerDestroyInstance( CPI_PrivateData	handle )
{
	CPI_Bool	 	 retval = CPI_FALSE;
	checkerState	*state	= (checkerState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}
	
	return retval; 
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

static CPI_Bool
CheckerProcess( CPI_PrivateData	 handle,
				CPI_Image		*result,
				CPI_Image	 	 /*inputs*/[],
				CPI_Uint32	 	 /*numInputs */ )
{
	CPI_Int32	 	 i, sizeY;
	CPI_Bool	 	 retval = CPI_FALSE;
	checkerState	*state	= (checkerState *)handle;
	
	sizeY = cpiRound( state->size[1] * state->scale[1] );
	
	if ( result->myContext.bitsPerPel == 8 )
	{
		for ( i = 0; i < result->myContext.channels; ++i )
		{
			state->fgColor[i] = cpiClampFloat( state->fgColor[i], 0.0, 1.0 );
			state->bgColor[i] = cpiClampFloat( state->bgColor[i], 0.0, 1.0 );
		}
		
		make_ramp( result, sizeY, state->squareSize, 
				   state->fgColor, state->bgColor, (CPI_Uint8)255 );
	}
	else if ( result->myContext.bitsPerPel == 16 )
	{
		for ( i = 0; i < result->myContext.channels; ++i )
		{
			state->fgColor[i] = cpiClampFloat( state->fgColor[i], 0.0, 1.0 );
			state->bgColor[i] = cpiClampFloat( state->bgColor[i], 0.0, 1.0 );
		}
		
		make_ramp( result, sizeY, state->squareSize, 
				   state->fgColor, state->bgColor, (CPI_Uint16)65535 );
	}
	else
		make_ramp( result, sizeY, state->squareSize, 
				   state->fgColor, state->bgColor, (CPI_Float32)1.0F );

	retval = CPI_TRUE;
	
	return retval;
}


//////////////////////////////////////
// This defines for Rayz the output region of your image
// op. It just passes on the values from the GUI.

static CPI_Bool
CheckerOutputRegion( CPI_PrivateData	 handle,
					 CPI_ImageContext	*output,
					 CPI_ImageContext	 /*inputs*/[],
					 CPI_Uint32			 /*numInputs*/ )
{	
	CPI_Bool 		 retval	= CPI_FALSE;
	checkerState	*state	= (checkerState *)handle;

	if ( NULL != state )
	{
		output->sizeX		= state->size[0];
		output->sizeY		= state->size[1];
		output->offsetX		= 0;
		output->offsetY		= 0;
		output->channels	= state->channels;
		output->bitsPerPel	= state->bitdepth;
		retval				= CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Checker parameter(s)" );
	
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"checker",			// op name
		"Checker Image Operation",
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	CheckerCreateInstance,	// create new checker imageop
	CheckerDestroyInstance,	// destroy checker imageop
	CheckerProcess,			// image execution process
	NULL,					// InputRegion
	CheckerOutputRegion,	// output region
	NULL,					// control processing
	NULL,					// setup processing
	NULL,					// cleanup processing
	0,						// min inputs
	CPI_FALSE,				// does it multiproc?
	CPI_FALSE				// handle disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "checker" );
	cpiUnregisterNode( "checker" );
}
