/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	CircleRamp
//			This is another source node, and this is shipped with RAYZ.
//			This shows overlays as well as nodes that generate an image
//			with no input image.
//
// COMMENTS:
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>
#include "CircleRamp.h"


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _rampState
{
	CPI_Int32	 channels;
	CPI_Int32	 bitdepth;
	CPI_Int32	 size[2];
	CPI_Int32	 center[2];
	CPI_Float32	 radius[2];
	CPI_Float32	 aspect;
	CPI_Float32	 edgeColors[4];
	CPI_Float32	 centerColors[4];
	CPI_Float32	 maxDim;
} rampState;


//////////////////////////////////////

static const char *channelmenu[]= 
{
	"RGB Alpha",
	"RGB only",
	"Alpha only",
	NULL
};


///////////////////////////////////////


static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////


static CPI_Float32 centerdefs[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 edgedefs[] = { 0.0F, 0.0F, 0.0F, 0.0F };


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
// Node Functions /////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static CPI_Bool
CircleRampInit( void )
{
	cpiAddColorGroup( "center_colors",
					  "Center Colors", 
					  centerdefs, 
					  4, 
					  CPI_PARM_DEFAULT,
					  "Adjust the color at the center of the ramp",
					  NULL );
	cpiAddColorGroup( "edge_colors",
					  "Edge Colors", 
					  edgedefs, 
					  4, 
					  CPI_PARM_DEFAULT,
					  "Adjust the color at the edges of the ramp",
					  NULL );

	cpiAddFloatPair( "center",
					 "Center", 
					 "x",
					 "X",
					 0.5, 
					 "y",
					 "Y",
					 0.5,
					 0.0,
					 1.0,
					 CPI_PARM_ANIMATABLE,
					 "Adjust the X/Y position of the center of the ramp",
					 NULL );
	cpiAddFloatPair( "radius",
					 "Radius", 
					 "in",
					 "In", 
					 0.0, 
					 "out",
					 "Out", 
					 1.0,
					 0.0,
					 1.0,
					 CPI_PARM_ANIMATABLE,
					 "Adjust the In/Out radius of the ramp",
					 NULL );
	cpiAddFloat( "aspect",
				 "Aspect", 
				 1.0, 
				 0.01, 
				 2, 
				 CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				 "Adjust the shape of the ramp",
				 NULL );
	cpiAddIntPair( "size",
				   "Size",
				   "width",
				   "Width", 
				   512, 
				   "height",
				   "Height", 
				   512,
				   1,
				   4096,
				   CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				   "Adjust the X/Y size of the output image",
				   NULL );
	cpiAddMenu( "bit_depth",
				"Bit Depth", 
				0, 
				bitdepthmenu,
				"Bit depth to be output",
				NULL );
	cpiAddMenu( "channels",
				"Channels", 
				0,
				channelmenu,
				"Number of channels to output",
				NULL );

	cpiAddSequenceRange( 1, 
						 1, 
						 CPI_PARM_DEFAULT, 
						 "Frame Numbers To Output",
						 NULL );

	// set up the node icon
	cpiRegisterIconData( "CircleRamp.tiff", CircleRamp_SIZEX, CircleRamp_SIZEY,
	  					  CircleRamp_data, 255, 0, 255 );
	// define the name of the checker icon
	cpiDefineIcon( "Nodes/CircleRamp", "CircleRamp.tiff",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );

	return CPI_TRUE;
}

///////////////////////////////////////////


static CPI_Bool
CircleRampShutDown( void )
{
	return CPI_TRUE;
}


///////////////////////////////////////////


static CPI_Bool
CircleRampGetRegion( CPI_PrivateData	 /*handle*/,
					 CPI_ImageContext	*output,
					 CPI_Float32		 myTime,
					 CPI_Uint32			 /*nodeOutput*/,
					 CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 startFrame <= myTime &&
		 endFrame >= myTime )
	{
		CPI_Int32	bitsPerPel, channels;
	
		if ( cpiGetInteger( &output->sizeX, "size.width", myTime ) &&
			 cpiGetInteger( &output->sizeY, "size.height", myTime ) &&
			 cpiGetInteger( &bitsPerPel, "bit_depth", myTime ) &&
			 cpiGetInteger( &channels, "channels", myTime ) )
		{
			output->offsetX = 0;
			output->offsetY = 0;

			switch ( bitsPerPel )
			{
				case 0:
					output->bitsPerPel = 8;
					break;
				case 1:
					output->bitsPerPel = 16;
					break;
				case 2:
					output->bitsPerPel = 32;
					break;
				default:
					cpiError( "Invalid bitsPerPel in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			switch ( channels )
			{
				case 0:
					output->channels = 4;
					break;
				case 1:
					output->channels = 3;
					break;
				case 2:
					output->channels = 1;
					break;
				default:
					cpiError( "Invalid channels in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			output->isPremultiplied = false;
		}
		else
		{
			cpiError( "Couldn't get parameter(s)" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Frame Out Of Range" );
		retval = CPI_FALSE;
	}
	
	return retval;
}


///////////////////////////////////////////


static void
CircleRampParmChanged( CPI_PrivateData	 /*handle*/,
					   const char 		*name )
{
	// Only sync with the global range when the node is first placed down
	if ( name == NULL )
	{
		CPI_Int32	 startFrame, endFrame;

		if ( cpiGetGlobalFrameRange( &startFrame, &endFrame ) )
			cpiSetSequenceParmRange( startFrame, endFrame );
	}
}


////////////////////////////////////////


static CPI_Bool
CircleRampGetFullSize( CPI_PrivateData	 /*handle*/,
					   CPI_ImageContext	*fullsize,
					   CPI_Float32		 myTime,
					   CPI_Uint32		 /*output*/,
					   CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool 	 retval = CPI_FALSE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( cpiGetInteger( &fullsize->sizeX, "size.width", myTime ) &&
			 cpiGetInteger( &fullsize->sizeY, "size.height", myTime ) )
			retval = CPI_TRUE;
		else
			cpiError( "Couldn't get full size parameters" );
	}
	else
		cpiError( "Frame Out Of Range" );
	
	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations


static CPI_ImageOp
CircleRampExec( CPI_PrivateData	 /*handle*/,
				CPI_Float32		 myTime,
				CPI_Uint8		 /*quality*/,
				CPI_Uint32		 /*output*/,
				CPI_Bool		 /*viewerOutput*/,
				CPI_Float32		 scaleX,
				CPI_Float32		 scaleY )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_ImageOp	 retval	= NULL;
	
	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		CPI_Int32	 	 channels, bitdepth;
		CPI_Float32	 	 centerX, centerY;	
		CPI_PrivateData	 opData;
		rampState		*rampData	 = NULL;

		// allocate a parameter structure of type 'circleramp'
		opData		= cpiCreatePrivateData( "circleramp" );
		rampData	= (rampState *)opData;

		if ( NULL != rampData )
		{
			// fill in the parm structure with user input values
			if ( cpiGetInteger( &rampData->size[0], "size.width", myTime ) &&
				 cpiGetInteger( &rampData->size[1], "size.height", myTime ) &&
				 cpiGetInteger( &channels, "channels", myTime ) &&
				 cpiGetInteger( &bitdepth, "bit_depth", myTime ) &&
				 cpiGetFloat( &centerX, "center.x", myTime ) &&
				 cpiGetFloat( &centerY, "center.y", myTime ) &&
				 cpiGetFloat( &rampData->radius[0], "radius.in", myTime ) &&
				 cpiGetFloat( &rampData->radius[1], "radius.out", myTime ) &&
				 cpiGetFloat( &rampData->aspect, "aspect", myTime ) &&
				 cpiGetFloat( &rampData->edgeColors[0],
							  "edge_colors.red", myTime ) &&
				 cpiGetFloat( &rampData->edgeColors[1],
							  "edge_colors.green", myTime ) &&
				 cpiGetFloat( &rampData->edgeColors[2],
							  "edge_colors.blue", myTime ) &&
				 cpiGetFloat( &rampData->edgeColors[3],
							  "edge_colors.alpha", myTime ) &&
				 cpiGetFloat( &rampData->centerColors[0],
							  "center_colors.red", myTime ) &&
				 cpiGetFloat( &rampData->centerColors[1],
							  "center_colors.green", myTime ) &&
				 cpiGetFloat( &rampData->centerColors[2],
							  "center_colors.blue", myTime ) &&
				 cpiGetFloat( &rampData->centerColors[3],
							  "center_colors.alpha", myTime ) )
			{
				switch ( channels )
				{
					case 0:
						rampData->channels = 4;
						break;
					case 1:
						rampData->channels = 3;
						break;
					case 2:
						rampData->channels = 1;
						break;
				}		

				switch ( bitdepth )
				{
					case 0:
						rampData->bitdepth = 8;
						break;
					case 1:
						rampData->bitdepth = 16;
						break;
					case 2:
						rampData->bitdepth = 32;
						break;
				}

				rampData->size[0]	= cpiRound( rampData->size[0] * scaleX );
				rampData->size[1]	= cpiRound( rampData->size[1] * scaleY );
				
				rampData->center[0]	= cpiRound( centerX * rampData->size[0] );
				rampData->center[1]	= cpiRound( ( 1.0 - centerY ) *
												rampData->size[1] );
				rampData->maxDim 	= ( 0.5 * ( ( rampData->size[0] >
												  rampData->size[1] )
												? rampData->size[1] :
												rampData->size[0] ) );

				// add the new IM to the exec list
				retval = cpiAddImageOp( "circleramp", rampData, NULL, 0 );
			}
			else
			{
				cpiError( "Couldn't get parameter(s)" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
			cpiError( "Couldn't get CircleRamp private data" );
	}
	else
		cpiError( "Frame Out Of Range" );
		
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not defined are set to NULL. Routines
// which are required to be provided are marked as such.

static RPI_NodeInfo ninfo = {
	{						// this structure is required
		"circleramp",
		"CircleRamp",
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	CircleRampInit,			// required
	CircleRampShutDown,
	NULL,		// CreateInstance
	NULL,		// DestroyInstance
	CircleRampParmChanged,
	NULL,		// InputAdded
	NULL,		// InputChanged
	NULL,		// InputRemoved
	CircleRampGetRegion,
    CircleRampGetFullSize,
	NULL,		// CircleRampGetRange,
	NULL,
	CircleRampExec,			// required
	NULL,		// InputLabel,
	NULL,		// OutputLabel,
	NULL,		// ViewerOutputLabel
	NULL,		// NumViewerOutputs
	0, 			// min inputs
	0,			// max inputs
	1,			// min outputs
	1,			// max outputs
	CPI_FALSE	// is growable (eg Mcomp)
};


//////////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.

template <class T> static
void make_ramp( CPI_Image	*image, 
				CPI_Float32	 radii[],
				CPI_Int32	 center[], 
			    CPI_Uint32	 chans,
				CPI_Float32	 aspect, 
				CPI_Float32	 cvals[], 
				CPI_Float32	 evals[],
				CPI_Float32	 maxdim,
				T 			 max )
{
	CPI_Uint32	 i, x, y;
	CPI_Float32	 dx, dy, d, xcen, ycen;
	CPI_Float32	 xscale, yscale;
	CPI_Float32	 minval=0.0, maxval=1.0;
	CPI_Float32  minrad=radii[0], maxrad=radii[1], scalefac;
	CPI_Uint32	 endX, endY;
	CPI_Uint8	 parmbase;
	T			*ptr, val;

	ptr		= (T *)image->data;
	xcen	= (CPI_Float32)center[0];
	ycen	= (CPI_Float32)center[1];
	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	// next 3 lines just for boundary conditions
	d = maxrad - minrad;
	if ( fabs( d ) < 0.0001 )
		minrad = maxrad - 0.0001;

	scalefac = ( maxval - minval ) / ( maxrad - minrad );
	xscale = yscale = 1.0;

	if ( aspect < 1.F )
	   	yscale = 1 / aspect;
	else
		xscale = aspect; 

	if ( chans == 1 )
		parmbase = 3;
	else
		parmbase = 0;

	for ( y = image->myContext.offsetY;
		  y < endY && !cpiIsProcessingAborted(); y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			// for each (x, y) compute d, the distance to the
			// center of the image
			dx = ( (CPI_Float32)x + 0.5F - xcen ) * xscale;
			dy = ( (CPI_Float32)y + 0.5F - ycen ) * yscale;

			d = sqrt( dx * dx + dy * dy ) / maxdim;

			// make it so that 0.0 <= d <= 1.0
			if ( d < minrad ) d = minrad;
			else if ( d > maxrad ) d = maxrad;
			d = ( d - minrad ) * scalefac + minval;
			
			// put in color, update pointer
			for ( i = 0; i < chans; i++ )
			{
				val = (T)( ( ( (1 - d) * cvals[i + parmbase] ) 
							 + (d * evals[i + parmbase]) ) * max );
				*ptr++ = val;
			}
		}
	}
}


/////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
CircleRampCreateOPInstance( void )
{
	rampState	*retval = (rampState *)cpiAlloc( sizeof( rampState ) );
	
	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.


static CPI_Bool
CircleRampDestroyOPInstance( CPI_PrivateData	handle )
{
	CPI_Bool	 retval 	= CPI_FALSE;
	rampState	*rampData	= (rampState *)handle;

	if ( NULL != rampData )
	{
		cpiFree( rampData );
		retval = CPI_TRUE;
	}

	return retval;	
}


////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth


static CPI_Bool
CircleRampProcess( CPI_PrivateData	 handle,
				   CPI_Image		*result,
				   CPI_Image	 	 /*inputs*/[],
				   CPI_Uint32	 	 /*numInputs*/ )
{
	CPI_Int32	i;
	CPI_Bool	retval		= CPI_FALSE;
	rampState	*rampData	= (rampState *)handle;

	if ( NULL != rampData )
	{
		if ( result->myContext.bitsPerPel == 8 )
		{
			for ( i = 0; i < result->myContext.channels; ++i )
			{
				rampData->centerColors[i] =
					cpiClampFloat( rampData->centerColors[i], 0.0, 1.0 );

				rampData->edgeColors[i] =
					cpiClampFloat( rampData->edgeColors[i], 0.0, 1.0 );
			}

			make_ramp( result, rampData->radius, rampData->center, 
					   result->myContext.channels, rampData->aspect,
					   rampData->centerColors, rampData->edgeColors,
					   rampData->maxDim, (CPI_Uint8)255 );
		}
		else if ( result->myContext.bitsPerPel == 16 )
		{
			for ( i = 0; i < result->myContext.channels; ++i )
			{
				rampData->centerColors[i] =
					cpiClampFloat( rampData->centerColors[i], 0.0, 1.0 );

				rampData->edgeColors[i] =
					cpiClampFloat( rampData->edgeColors[i], 0.0, 1.0 );
			}
			
			make_ramp( result, rampData->radius, rampData->center, 
					   result->myContext.channels, rampData->aspect,
					   rampData->centerColors, rampData->edgeColors,
					   rampData->maxDim, (CPI_Uint16)65535 );
		}
		else if ( result->myContext.bitsPerPel == 32 )
			make_ramp( result, rampData->radius, rampData->center, 
					   result->myContext.channels, rampData->aspect,
					   rampData->centerColors, rampData->edgeColors,
					   rampData->maxDim, (CPI_Float32)1.0F );
		else
			cpiError( "Result image has an invalid bit depth!" );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Circleramp couldn't get private data" );

	return retval;
}


//////////////////////////////////////


static CPI_Bool
CircleRampControlProcessing( CPI_PrivateData	 /*handle*/,
							 CPI_Image			*/*result*/,
							 CPI_Image	 	 	 /*inputs*/[],
							 CPI_Uint32	 	 	 /*numInputs*/ )
{
	CPI_Bool	retval = CPI_TRUE;

	retval = cpiCompute( true );

	return retval;
}


//////////////////////////////////////


static CPI_Bool
CircleRampOutputRegion( CPI_PrivateData		 handle,
						CPI_ImageContext	*output,
						CPI_ImageContext	 /*inputs*/[],
						CPI_Uint32			 /*numInputs*/ )
{	
	CPI_Bool 	 retval		= CPI_FALSE;
	rampState	*rampData	= (rampState *)handle;

	if ( NULL != rampData )
	{
		output->sizeX		= rampData->size[0];
		output->sizeY		= rampData->size[1];
		output->bitsPerPel	= rampData->bitdepth;
		output->channels	= rampData->channels;
		output->offsetX 	= 0;
		output->offsetY 	= 0;
		retval				= CPI_TRUE;
	}
	else
		cpiError( "Circleramp got NULL privateData" );
	
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.
// Required entries are marked, the rest are optional. Fill in
// NULL if there is no entry for an optional function.


static RPI_ImageOpInfo opinfo = 
{
	{						// this structure is required
		"circleramp",
		"CircleRamp Image Operation",
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	CircleRampCreateOPInstance,	
	CircleRampDestroyOPInstance,
	CircleRampProcess,		// required
	NULL,					//CircleRampInputRegion
	CircleRampOutputRegion,
	CircleRampControlProcessing,
	NULL,					//CircleRampSetupProcessing
	NULL,					//CircleRampCleanupProcessing
	0,						// number of inputs required
	CPI_TRUE,				// can this multiproc?
	CPI_FALSE,				// disjoint regions?
};


///////////////////////////////////////////////////////////////////////////////
// Overlay Functions //////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

// The overlays state has a similar structure and allocation/freeing
// mechanism to the node's parameter structure. Here we define the
// structure, then the routines to allocate and free it.

#define INNER_CIRCLE_ID		1
#define OUTER_CIRCLE_ID		2
#define CROSS_ID			3

typedef struct _overlayState
{
	CPI_Int32	fullSizeX;
	CPI_Int32	fullSizeY;
	CPI_Float32	aspect;
	CPI_Float32	centerX;
	CPI_Float32 centerY;
	CPI_Float32	radiusIn;
	CPI_Float32 radiusOut;
	CPI_Float32	selectedX;
	CPI_Float32 selectedY;
	CPI_Uint32	selectedObject;
	CPI_Float32	myTime;
} overlayState;


//////////////////////////////////////


static CPI_PrivateData
CircleRampCreateInstance( void )
{
	overlayState	*retval;

	retval = (overlayState *)cpiAlloc( sizeof( overlayState ) );
	
	return retval;
}


///////////////////////////////////////


static void
CircleRampDestroyInstance( CPI_PrivateData handle )
{
	overlayState *state = (overlayState *)handle;

	cpiFree( state );
}


///////////////////////////////////////
// Define the routine to actually draw (render) the overlay on
// the screen


static void
CircleRampRender( CPI_PrivateData 	handle, 
				  CPI_Bool 			isPicking )
{
	CPI_Float32	  	 inRad, outRad, imageAspectX, imageAspectY;
	overlayState	*state	= (overlayState *)handle;
	CPI_Bool		 canSet	= CPI_FALSE;
	
	inRad = state->radiusIn * 0.5;
	outRad = state->radiusOut * 0.5;

	if ( inRad <= 0 )
		inRad = 0.075;
	
	if ( outRad <= 0 )
		outRad = 0.1;

	if ( state->fullSizeX > state->fullSizeY )
	{
		imageAspectX	= ( state->fullSizeY * inRad ) / state->fullSizeX;
		imageAspectY	= inRad;
	}
	else if ( state->fullSizeX < state->fullSizeY )
	{
		imageAspectX	= inRad;
		imageAspectY	= ( state->fullSizeX * inRad ) / state->fullSizeY;
	}
	else
	{
		imageAspectX	= inRad;
		imageAspectY	= inRad;
	}

	imageAspectX *= ( 1 / state->aspect );
	imageAspectY *= state->aspect;

	canSet = cpiCanSetParameter( "radius.in", state->myTime );

	cpiSetColor( 0.0, 0.0, 1.0 );

	if ( canSet )
	{		
		cpiStartNewObject( INNER_CIRCLE_ID );
		cpiDrawEllipse( state->centerX, 
						state->centerY, 
						imageAspectX, 
						imageAspectY );
	}
	else
	{
		if ( !isPicking )
		{
			cpiStartLineStyle( CPI_LINE_STYLE_DASH_SHORT );
			cpiStartNewObject( INNER_CIRCLE_ID );
			cpiDrawEllipse( state->centerX, 
							state->centerY, 
							imageAspectX, 
							imageAspectY );
			cpiEndLineStyle();
		}
	}
	
	if ( state->fullSizeX > state->fullSizeY )
	{
		imageAspectX	= ( state->fullSizeY * outRad ) / state->fullSizeX;
		imageAspectY	= outRad;
	}
	else if ( state->fullSizeX < state->fullSizeY )
	{
		imageAspectX	= outRad;
		imageAspectY	= ( state->fullSizeX * outRad ) / state->fullSizeY;
	}
	else
	{
		imageAspectX	= outRad;
		imageAspectY	= outRad;
	}

	imageAspectX *= ( 1 /state->aspect );
	imageAspectY *= state->aspect;

	canSet = cpiCanSetParameter( "radius.out", state->myTime );

	if ( canSet )
	{
		cpiStartNewObject( OUTER_CIRCLE_ID );
		cpiDrawEllipse( state->centerX, 
						state->centerY, 
						imageAspectX, 
						imageAspectY );
	}
	else
	{
		if ( !isPicking )
		{
			cpiStartLineStyle( CPI_LINE_STYLE_DASH_SHORT );
			cpiStartNewObject( OUTER_CIRCLE_ID );
			cpiDrawEllipse( state->centerX, 
							state->centerY, 
							imageAspectX, 
							imageAspectY );
			cpiEndLineStyle();
		}
	}

	canSet = ( cpiCanSetParameter( "center.x", state->myTime ) &&
			   cpiCanSetParameter( "center.y", state->myTime ) ); 

	if ( canSet )
	{
		cpiStartNewObject( CROSS_ID );
		cpiDrawLine( state->centerX - 0.05, 
					 state->centerY, 
					 state->centerX + 0.05, 
					 state->centerY );
		cpiDrawLine( state->centerX, 
					 state->centerY - 0.05,
					 state->centerX, 
					 state->centerY + 0.05 );
	}
	else
	{
		if ( !isPicking )
		{
			cpiStartLineStyle( CPI_LINE_STYLE_DASH_SHORT );
			cpiStartNewObject( CROSS_ID );
			cpiDrawLine( state->centerX - 0.05, 
						 state->centerY, 
						 state->centerX + 0.05, 
						 state->centerY );
			cpiDrawLine( state->centerX, 
						 state->centerY - 0.05,
						 state->centerX, 
						 state->centerY + 0.05 );
			cpiEndLineStyle();
		}
	}
}


//////////////////////////////////////////
// This is normally called when the input image is replaced. In the
// case of a node with no input image, we get here when the full
// size changes


static void
CircleRampSourceOutOfDate( CPI_PrivateData 	handle, 
						   CPI_ImageContext	fullsize,
						   CPI_Float32 		theTime )
{
	overlayState *state = (overlayState *)handle;

	cpiGetFloat( &state->centerX, "center.x", theTime );
	cpiGetFloat( &state->centerY, "center.y", theTime );
	cpiGetFloat( &state->radiusIn, "radius.in", theTime );
	cpiGetFloat( &state->radiusOut, "radius.out", theTime );
	cpiGetFloat( &state->aspect, "aspect", theTime );
	
	state->fullSizeX = fullsize.sizeX;
	state->fullSizeY = fullsize.sizeY;

	state->myTime = theTime;
}


///////////////////////////////////////////


static void
CircleRampLeftMouseDown( CPI_PrivateData	handle,
						 CPI_Uint32			objectID,
						 CPI_Float32		x,
						 CPI_Float32		y,
						 CPI_Uint32			/*keyMod*/ )
{
	overlayState	*state = (overlayState *)handle;

	state->selectedObject	= objectID;

	state->selectedX		= x;
	state->selectedY		= y;
	
	cpiSetSelectLock( CPI_TRUE );	
}


/////////////////////////////////////////


static void
CircleRampLeftMouseDrag( CPI_PrivateData	 handle,
						 CPI_Uint32			 /*objectID*/,
						 CPI_Float32		 x,
						 CPI_Float32		 y,
						 CPI_Uint32			 /*keyMod*/,
						 CPI_Bool			 isContinuous )
{
	overlayState	*state = (overlayState *)handle;

	if ( state->selectedObject == CROSS_ID )
	{
		state->centerX = x;
		state->centerY = y;
	}
	else if ( state->selectedObject == OUTER_CIRCLE_ID )
	{
		state->radiusOut = ( hypot( x - state->centerX, y - state->centerY ) *
							 2 );
	}
	else if ( state->selectedObject == INNER_CIRCLE_ID )
	{
		state->radiusIn = ( hypot( x - state->centerX, y - state->centerY ) * 
							2 );
	}

	if ( isContinuous )
	{
		cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
		cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE  );
		cpiSetFloat( state->radiusOut, "radius.out", state->myTime, CPI_TRUE );
		cpiSetFloat( state->radiusIn, "radius.in", state->myTime, CPI_TRUE );
	}
	
	cpiDoRedraw();   
}


/////////////////////////////////////////


static void
CircleRampLeftMouseUp( CPI_PrivateData	 handle, 
					   CPI_Uint32		 objectID,
					   CPI_Float32		 /*x*/,
					   CPI_Float32		 /*y*/,
					   CPI_Uint32		 /*keyMod*/ )
{
	overlayState	*state = (overlayState *)handle;
	
	cpiSetSelectLock( CPI_FALSE );

	if ( state->selectedObject == CROSS_ID )
	{
		cpiSetFloat( state->centerX, "center.x", state->myTime, CPI_TRUE );
		cpiSetFloat( state->centerY, "center.y", state->myTime, CPI_TRUE  );
	}
	else if ( state->selectedObject == OUTER_CIRCLE_ID )
		cpiSetFloat( state->radiusOut, "radius.out", state->myTime, CPI_TRUE );
	else if ( state->selectedObject == INNER_CIRCLE_ID )
		cpiSetFloat( state->radiusIn, "radius.in", state->myTime, CPI_TRUE );
}


////////////////////////////////////////
// The overlay routines are also held in a structure
// The required function is marked; optional functions 
// can be listed as NULL when not required


static RPI_OverlayInfo oinfo = {
	{
		"circleramp",
		"CircleRamp",
		"Silicon Grail",
		"1.0", NULL, NULL
	},
	NULL,		// Init
	NULL,		// Shutdown
	CircleRampCreateInstance,
	CircleRampDestroyInstance,
	CircleRampSourceOutOfDate,
	NULL,		// CircleRampFullSizeChanged,
	NULL,		// CircleRampParmChanged
	CircleRampSourceOutOfDate,
	CircleRampRender,			// (required)
	CircleRampLeftMouseDown,
	CircleRampLeftMouseDrag,
	CircleRampLeftMouseUp,
	NULL,		// CircleRampMouseEnter,
	NULL,		// CircleRampMouseDrag,
	NULL,		// CircleRampMouseExit,
	// this attaches the overlay to the named node. If you want
	// the overlay to appear in the Image Viewer for general use,
	// pass in NULL for the node name.
	"circleramp",
	// if the following is CPI_TRUE, you will receive mouse events even
	// when the mouse is not over any object that you have previously
	// rendered. In that case, you will get an event for object 0, which
	// is a reserved object ID. This applies, in, for instance, the roto
	// node, where you want to get mouse events to draw new shapes, not
	// just find existing ones.
	CPI_FALSE
};


//////////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = true;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	if ( retval )
		retval = cpiRegisterOverlay( &oinfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "circleramp" );
	cpiUnregisterNode( "circleramp" ); 
	cpiUnregisterOverlay( "circleramp" );
}
