/*
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Clamp
//
// COMMENTS: This clamps the rgb values of an image
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>
#include "Clamp.h"

#define MAX8	(CPI_Uint8)255
#define MAX16	(CPI_Uint16)65535


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _clampData
{
	CPI_Float32	 min[4];
	CPI_Float32	 max[4];
} clampState;


/////////////////////////////////////////
// Init is called when the node is created and is 
// typically used to define the user input controls.
// Here, there is nothing to do.

static CPI_Bool
ClampInit( void )
{
	cpiAddFloatRGBAOGroup( "minimum_list", "Min Value",
	                        0.0, 0.0, 1.0, 4, CPI_PARM_ANIMATABLE,
	                       "Minimum color values", NULL );

	cpiAddFloatRGBAOGroup( "maximum_list", "Max Value",
	                        1.0, 0.0, 1.0, 4, CPI_PARM_ANIMATABLE,
	                       "Maximum color values", NULL );

	// set up the node icon
	cpiRegisterIconData( "Clamp.tif", Clamp_SIZEX, Clamp_SIZEY,
	  					   Clamp_data, 255, 0, 255 );
	// define the name of the checker icon
	cpiDefineIcon( "Nodes/Clamp", "Clamp.tif",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed
// This gets the input image, and then adds the image
// operation to the list of ops to be executed

static CPI_ImageOp
ClampExec( CPI_PrivateData	/*handle*/,
		   CPI_Float32		myTime,
		   CPI_Uint8		quality,
		   CPI_Uint32		nodeOutput,
		   CPI_Bool	 		viewerOutput,
		   CPI_Float32		scaleX,
		   CPI_Float32   	scaleY )

{
	CPI_ImageOp 	 myInputs;
	CPI_PrivateData	 opData;
	clampState		*opState	= NULL;
	CPI_ImageOp		 retval		= NULL;

	// Allocate a parameter structure as previously defined
	opData		= cpiCreatePrivateData( "clamp" );
	opState		= (clampState *)opData;

	if ( NULL != opState )
	{
		// get the clamp values from the user and put them into the
		// parm structure
		// Note: don't need to get or use the overall value for each
		// group, because that is already factored in to the rgba values
		if ( cpiGetFloat( &opState->min[0], "minimum_list.red", myTime ) &&
			 cpiGetFloat( &opState->min[1], "minimum_list.green", myTime ) &&
			 cpiGetFloat( &opState->min[2], "minimum_list.blue", myTime ) &&
			 cpiGetFloat( &opState->min[3], "minimum_list.alpha", myTime ) &&
			 cpiGetFloat( &opState->max[0], "maximum_list.red", myTime ) &&
			 cpiGetFloat( &opState->max[1], "maximum_list.green", myTime ) &&
			 cpiGetFloat( &opState->max[2], "maximum_list.blue", myTime ) &&
			 cpiGetFloat( &opState->max[3], "maximum_list.alpha", myTime ) )
		{
			myInputs	= cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

			if ( NULL != myInputs )
				retval = cpiAddImageOp( "clamp", opState, &myInputs, 1 );
			else
				cpiError( "Missing required input" );
		}
		else
			cpiError( "Couldn't get Clamp parameter(s)" );
	}
	else
		cpiError( "Couldn't get Clamp private data" );
	
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"clamp",				// node name
		"Clamp",				// menu name
		"Joe Programmer",		// author
		"2.2"					// version
	},
	ClampInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// CreateInstance
	NULL,				// DestroyInstance
	NULL,				// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	ClampExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,               // ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.
// This just copies input to result

template <class T> static
void clamp( CPI_Image	&input,
			CPI_Image	*result, 
			T 			 min[],
			T			 max[] )
{
	CPI_Int32	 x, y, endX, endY, c;
	CPI_Int32	 channels = result->myContext.channels;
	T			*pin, *pout, p;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			for ( c = 0; c < channels; c++ )
			{
				/* Threshold the pixel */
				p = *pin;
				if ( p < min[c] )
					p = min[c];
				else if (p > max[c] )
					p = max[c];
				*pout = p;

				pin++;
				pout++;
			}
		}
	}
}


////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
ClampCreateOpInstance( void )
{
	clampState	*retval = (clampState *)cpiAlloc( sizeof( clampState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.


static CPI_Bool
ClampDestroyOpInstance( CPI_PrivateData handle )
{
	CPI_Bool	 retval = CPI_FALSE;
	clampState	*state	= (clampState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called to execute the image op
// It just calls the appropriate copy routine, 
// based on the input bit depth

static CPI_Bool
ClampProcess( CPI_PrivateData	 handle,
			  CPI_Image			*result,
			  CPI_Image	 	 	 inputs[],
			  CPI_Uint32	 	 numInputs )
{
	CPI_Int32	 i;
	CPI_Bool	 retval = CPI_FALSE;
	clampState	*state	= (clampState *)handle;

	if ( NULL != state )
	{		
		// set up to call appropriate routine
		// no bounds checking or sanity checks on values
		if ( result->myContext.bitsPerPel == 8 )
		{
			CPI_Uint8	min[10], max[10];
		
			min[0] = cpiRound( state->min[0] * MAX8 );
			min[1] = cpiRound( state->min[1] * MAX8 );
			min[2] = cpiRound( state->min[2] * MAX8 );
			min[3] = cpiRound( state->min[3] * MAX8 );
		
			max[0] = cpiRound( state->max[0] * MAX8 );
			max[1] = cpiRound( state->max[1] * MAX8 );
			max[2] = cpiRound( state->max[2] * MAX8 );
			max[3] = cpiRound( state->max[3] * MAX8 );
			for ( i = 4; i < 10; i++)
			{
				min[i] = (CPI_Uint8)0;
				max[i] = MAX8;
			}

			clamp( inputs[0], result, min, max );
		}
		else if ( result->myContext.bitsPerPel == 16 )
		{
			CPI_Uint16	min[10], max[10];

			min[0] = cpiRound( state->min[0] * MAX16 );
			min[1] = cpiRound( state->min[1] * MAX16 );
			min[2] = cpiRound( state->min[2] * MAX16 );
			min[3] = cpiRound( state->min[3] * MAX16 );

			max[0] = cpiRound( state->max[0] * MAX16 );
			max[1] = cpiRound( state->max[1] * MAX16 );
			max[2] = cpiRound( state->max[2] * MAX16 );
			max[3] = cpiRound( state->max[3] * MAX16 );
			for ( i = 4; i < 10; i++)
			{
				min[i] = (CPI_Uint16)0;
				max[i] = MAX16;
			}

			clamp( inputs[0], result, min, max );
		}
		else
		{
			CPI_Float32	min[10], max[10];

			min[0] = state->min[0];
			min[1] = state->min[1];
			min[2] = state->min[2];
			min[3] = state->min[3];

			max[0] = state->max[0];
			max[1] = state->max[1];
			max[2] = state->max[2];
			max[3] = state->max[3];
			for ( i = 4; i < 10; i++)
			{
				min[i] = 0.0F;
				max[i] = 1.0F;
			}

			clamp( inputs[0], result, min, max );
		}

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get clamp parameter(s)" );

	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM

static RPI_ImageOpInfo opinfo = 
{
	{
		"clamp",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"2.2"				// version
	},
	ClampCreateOpInstance,	
	ClampDestroyOpInstance,	
	ClampProcess,			// image execution process
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE				// Disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "clamp" );
	cpiUnregisterNode( "clamp" );
}
