/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Prefs
//
// COMMENTS: Demonstrates setting RAYZ preference values
//
*/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <CPI.h>

/*
	This node shows two things. First, it shows how to create new, 
	custom preferences. You can do this via a node, even though the
	node is never intended to be used, because all nodes have their
	Init function called when RAYZ starts up. So in that Init(),
	you can register a callback for when a project is created, and
	when that happens, add any number of custom preferences.

	This can be registered in any Init() function: node, overlay, LUT, etc.

	The second thing demonstrated here is the setting of a preference
	value to be used in a calculation. When this node executes, it
	first gets the color value from the color wheel widget, then puts
	that color into the preference that controls the background color
	in the image viewer.
*/


/////////////////////////////////////////
// This function is called when a new project is created. RAYZ knows to find
// this function because it is registered in the PrefInit() function below.

static
void
newProjectHandler( const char *projName )
{
	if ( NULL != projName && projName[0] != '\0' )
	{
		// define and allocate a string to put the new path
		// name for the new preference into
		int		 projLen	= strlen( projName ) + 15;
		char	*newGroup	= (char *)cpiAlloc( projLen );

		if ( NULL != newGroup )
		{
			strcpy( newGroup, projName );
			strcat( newGroup, "/Color Space" );
		
			// add a custom preference to the Project Settings dialog
			cpiAddPreferenceGroup( newGroup );
			cpiAddPreferenceValue( newGroup,				// path
								   "Name",					// pref name
								   "Which color space to use",	// help
								   "RGB",					// default val
								   CPI_PREF_STRING,			// type
								   CPI_PREF_UI_DEFAULT );	// flags

			cpiFree( newGroup );
		}
	}
}

/////////////////////////////////////////
// Init is called when the node is created and is
// typically used to define the user input controls.

static CPI_Bool
PrefsInit( void )
{
	CPI_Float32     def[4] = {0.0, 0.0, 0.0, 1.0};

	cpiAddDisplayString("null", "", "Adjust image viewer background color" );
	cpiAddColorGroup("color", "Color", def, 4, CPI_PARM_DEFAULT,
			                 "Choose a new outside color", NULL );

	// This call tells RAYZ to call the given function whenever a
	// new project is created.
	cpiRegisterNewProjectHandler( newProjectHandler );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed
// This gets the input image, and then adds the image
// operation to the list of ops to be executed

#define OUTSIDE_COLOR \
	"User Interface/Settings/Image Viewer/Outside Image Color"

static CPI_ImageOp
PrefsExec( CPI_PrivateData	/*handle*/,
		   CPI_Float32		myTime,
		   CPI_Uint8		quality,
		   CPI_Uint32		nodeOutput,
		   CPI_Bool			viewerOutput,
		   CPI_Float32		scaleX,
		   CPI_Float32  	scaleY )

{
	CPI_ImageOp	retval	= NULL;

	// get the user's color choice from the color widget,
	// then set it into the Outside Image preference
	CPI_Float32		red, green, blue;

	cpiGetFloat( &red,   "color.red",   myTime );
	cpiGetFloat( &green, "color.green", myTime );
	cpiGetFloat( &blue,  "color.blue",  myTime );

	// This should cause the image viewer to change it's background pref based
	// on the color selected here. Yes, a silly example, but effective... :)
	cpiSetPreferenceColor( OUTSIDE_COLOR, red, green, blue );

	// all we do is pass the input to the output
	retval = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"preferences",			// node name
		"Preferences",			// menu name
		"Joe Programmer",		// author
		"2.2"					// version
	},
	PrefsInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// CreateInstance
	NULL,				// DestroyInstance
	NULL,				// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	PrefsExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,               // ViewerOutputs
	1,					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval = cpiRegisterNode( &ninfo );

	return retval;
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterNode( "preferences" );
}
