/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Ramp
//
// COMMENTS: This illustrates a source node (one with no inputs)
//			 Generates horizontal or vertical color ramps
//
*/

#include <stdio.h>
#include <CPI.h>
#include "Ramp.h"

#define MAX8       (CPI_Uint8)255
#define MAX16      (CPI_Uint16)65535
#define MAXF       (CPI_Float32)1.0F
#define VERTICAL	0
#define HORIZONTAL	1


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _rampState
{
	CPI_Int32	 size[2];
	CPI_Int32	 direction;
	CPI_Int32	 channels;
	CPI_Int32	 bitdepth;
	CPI_Float32	 fromColor[4];
	CPI_Float32	 toColor[4];
} rampState;


//////////////////////////////////////
// menu of which channels to create

static const char *channelmenu[]= 
{
	"RGB Alpha",
	"RGB",
	"Alpha only",
	NULL
};


///////////////////////////////////////
// menu of which bit depth to create

static const char *bitdepthmenu[] = 
{
	"8 bit",
	"16 bit",
	"Float",
	NULL
};


///////////////////////////////////////
// type of ramp

static const char *rampmenu[] = 
{
	"Vertical",
	"Horizontal",
	NULL
};


///////////////////////////////////////
// default color arrays

static CPI_Float32 fromC[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32 toC[]   = { 0.0F, 0.0F, 0.0F, 1.0F };


/////////////////////////////////////////////////////
// Node Functions ///////////////////////////////////
/////////////////////////////////////////////////////

// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
RampInit( void )
{
	cpiAddColorGroup( "fromC", "From Color", fromC, 4, CPI_PARM_DEFAULT, 
						"Color at start of ramp", NULL );
	cpiAddColorGroup( "toC", "To Color", toC, 4, CPI_PARM_DEFAULT, 
						"Color at end of ramp", NULL );
	cpiAddMenu( "direction", "Ramp is", 0, rampmenu, NULL, NULL );
	cpiAddIntPair( "size", "Size",
				   "w", "Width", 720,
				   "h", "Height", 576,
				   1, 4096, CPI_PARM_CLAMPMIN,
				   "Size of output image", NULL );
	cpiAddMenu( "bits", "Bit Depth", 0, bitdepthmenu, NULL, NULL );
	cpiAddMenu( "chans", "Channels", 0, channelmenu, NULL, NULL );
	cpiAddSequenceRange( 1, 100,
						CPI_PARM_DEFAULT,
						"Range of frames to output", NULL );

    // set up the ramp icon
    cpiRegisterIconData( "Ramp.tiff", 60, 45, Ramp,
                          255, 0, 255 );

    // define the location of the ramp icon
    cpiDefineIcon( "Nodes/Ramp", "Ramp.tiff",
                   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );

	return CPI_TRUE;
}


///////////////////////////////////////////
// GetRegion needs to be defined for nodes that generate their
// own images (as opposed to being given images to modify)
// This defines for the rest of Rayz what the output 'context'
// is, ie, size, bit depth, etc.

static CPI_Bool
RampGetRegion( CPI_PrivateData	 /*handle*/,
			   CPI_ImageContext	*output,
			   CPI_Float32		 myTime,
			   CPI_Uint32		 /*nodeOutput*/,
			   CPI_Bool		 	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		CPI_Int32	 bitsPerPel, channels;

		// get size from GUI and store in 'output'
		retval = ( cpiGetInteger( &output->sizeX, "size.w", myTime ) &&
				   cpiGetInteger( &output->sizeY, "size.h", myTime ) &&
				   cpiGetInteger( &bitsPerPel, "bits", myTime ) &&
				   cpiGetInteger( &channels, "chans", myTime ) );

		if ( retval )
		{
			output->offsetX = 0;
			output->offsetY = 0;

			switch ( bitsPerPel )
			{
				case 0:
					output->bitsPerPel = 8;
					break;
				case 1:
					output->bitsPerPel = 16;
					break;
				case 2:
					output->bitsPerPel = 32;	// ie, float
					break;
				default:
					cpiError( "Invalid bitsPerPel in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}

			// get user menu choice, and set value accordingly
			switch ( channels )
			{
				case 0:
					output->channels = 4;
					break;
				case 1:
					output->channels = 3;
					break;
				case 2:
					output->channels = 1;
					break;
				default:
					cpiError( "Invalid channels in getRegion!\n" );
					retval = CPI_FALSE;
					break;
			}
		}
		else
		{
			cpiError( "Couldn't get parameter!" );
			retval = CPI_FALSE;
		}
	}
	else
	{
		cpiError( "Frame Out of Range!" );
		retval = CPI_FALSE;
	}
	
	return retval;
}



////////////////////////////////////////


static void
RampParmChanged( CPI_PrivateData	 /*handle*/,
				 const char 		*parmname )
{
	// If the parm name is NULL, that means the a new node has just been added
	// we want to set the seuqence parm to match the global frame range.
	if ( parmname == NULL )
	{
		CPI_Int32	 startFrame, endFrame;

		if ( cpiGetGlobalFrameRange( &startFrame, &endFrame ) )
			cpiSetSequenceParmRange( startFrame, endFrame );
	}
}


///////////////////////////////////////////
// Rayz may need to ask the plugin what size
// image it intends to create - define that here
// Nodes that create their own output need to
// define this routine

static CPI_Bool
RampGetFullSize( CPI_PrivateData	 /*handle*/,
				 CPI_ImageContext	*fullsize,
				 CPI_Float32		 myTime,
				 CPI_Uint32		  	 /*nodeOutput*/,
				 CPI_Bool       	 /*viewerOutput*/ )
{
	CPI_Int32	 startFrame, endFrame;
	CPI_Bool 	 retval = CPI_TRUE;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		if ( !cpiGetInteger( &fullsize->sizeX, "size.w", myTime ) ||
			 !cpiGetInteger( &fullsize->sizeY, "size.h", myTime ) )
		{
			cpiError( "Couldn't get size parameter" );
			retval = CPI_FALSE;
		}			
	}
	else
	{
		cpiError( "Frame Out Of Range" );
		retval = CPI_FALSE;
	}		
		
	return retval;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the parameters from the user interface
// 2. allocate metadata and copy the parameters to it
// 3. add the image op(s) to the list of operations

static CPI_ImageOp
RampExec( CPI_PrivateData	 /*handle*/,
		  CPI_Float32 	 	 myTime,
		  CPI_Uint8	 	 	 /*quality*/,
		  CPI_Uint32 	 	 /*output*/,
		  CPI_Bool 	 	 	 /*viewerOutput*/,
		  CPI_Float32 	 	 scaleX,
		  CPI_Float32	 	 scaleY )
{
	CPI_Int32	startFrame, endFrame;
	CPI_Int32	channels, bitdepth;
	CPI_ImageOp	retval	= NULL;

	if ( cpiGetSequenceParmRange( &startFrame, &endFrame ) &&
		 myTime >= startFrame &&
		 myTime <= endFrame )
	{
		// These 4 lines define and allocate a block of data as
		// defined in the earlier typedef
		CPI_PrivateData	 opData;
		rampState		*rampData	= NULL;

		opData 		= cpiCreatePrivateData( "ramp" );
		rampData	= (rampState *)opData;

		if ( NULL != rampData )
		{
			// get the user parameter value and stuff it into the
            // structure. This structure is then added, along with
            // the input(s) and the image op itself, to the list
            // of operations that RAYZ will execute.
			if ( cpiGetInteger( &rampData->size[0], "size.w", myTime ) &&
				 cpiGetInteger( &rampData->size[1], "size.h", myTime ) &&
				 cpiGetInteger( &channels, "chans", myTime ) &&
				 cpiGetInteger( &bitdepth, "bits", myTime ) &&
				 cpiGetInteger( &rampData->direction, "direction", myTime ) &&
				 cpiGetFloat( &rampData->fromColor[0],
							  "fromC.red", myTime ) &&
				 cpiGetFloat( &rampData->fromColor[1],
							  "fromC.green", myTime ) &&
				 cpiGetFloat( &rampData->fromColor[2],
							  "fromC.blue", myTime ) &&
				 cpiGetFloat( &rampData->fromColor[3],
							  "fromC.alpha", myTime ) &&
				 cpiGetFloat( &rampData->toColor[0], "toC.red", myTime ) &&
				 cpiGetFloat( &rampData->toColor[1], "toC.green", myTime ) &&
				 cpiGetFloat( &rampData->toColor[2], "toC.blue", myTime ) &&
				 cpiGetFloat( &rampData->toColor[3], "toC.alpha", myTime ) )
			{
				switch ( channels )
				{
					case 0:
						rampData->channels	= 4;
						break;
					case 1:
						rampData->channels	= 3;
						break;
					case 2:
						rampData->channels	= 1;
						break;
				}

				switch ( bitdepth )
				{
					case 0:
						rampData->bitdepth	= 8;
						break;
					case 1:
						rampData->bitdepth	= 16;
						break;
					case 2:
						rampData->bitdepth	= 32;
						break;
				}

				retval = cpiAddImageOp( "ramp", rampData, NULL, 0 );
			}
			else
			{
				cpiError( "Couldn't get Ramp parameter(s)" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
			cpiError( "Couldn't get Ramp private data" );
	}
	else
		cpiError( "Frame Out Of Range" );
		
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not defined are set to NULL. Routines
// which are required to be provided are marked as such.

static RPI_NodeInfo ninfo = {
	{					// this structure is required
		"ramp",			// node name
		"Ramp",			// menu name
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	RampInit,			// NodeInit 	(required)
	NULL,				// NodeShutDown
	NULL,				// CreateInstance
	NULL,				// DestroyInstance
	RampParmChanged, 	// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	RampGetRegion,		// GetRegion
    RampGetFullSize,	// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	RampExec,			// Node Execute	(required)
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	0, 					// min input
	0,					// max input
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable	(eg Mcomp)
};


////////////////////////////////////////////////////
// Image Op Functions //////////////////////////////
////////////////////////////////////////////////////


////////////////////////////////////////
// This set of functions does the actual work. There is one
// for each direction for each possible number of channels
// to be output. In this way we unroll the looping over the
// number of channels, which saves about 10% of the processing
// time

// handles the vertical case of alpha-only

template <class T> static
void make_ramp_y_1( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col, delta;
	T			*ptr;

	// init color stuff
	col = fvals[3] * (CPI_Float32)max;
	tvals[3] *= (CPI_Float32)max;
	delta = ( tvals[3] - col ) / image->myContext.sizeY;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col;
		}

		col += delta;
	}
}

// handles the vertical rgb case

template <class T> static
void make_ramp_y_3( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col[3], delta[3];
	T			*ptr;

	// init color stuff
	col[0] = fvals[0] * (CPI_Float32)max;
	col[1] = fvals[1] * (CPI_Float32)max;
	col[2] = fvals[2] * (CPI_Float32)max;
	tvals[0] *= (CPI_Float32)max;
	tvals[1] *= (CPI_Float32)max;
	tvals[2] *= (CPI_Float32)max;
	delta[0] = ( tvals[0] - col[0] ) / image->myContext.sizeY;
	delta[1] = ( tvals[1] - col[1] ) / image->myContext.sizeY;
	delta[2] = ( tvals[2] - col[2] ) / image->myContext.sizeY;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col[0];
			*ptr++ = (T)col[1];
			*ptr++ = (T)col[2];
		}

		col[0] += delta[0];
		col[1] += delta[1];
		col[2] += delta[2];
	}
}

// handles the vertical rgba case

template <class T> static
void make_ramp_y_4( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col[4], delta[4];
	T			*ptr;

	// init color stuff
	col[0] = fvals[0] * (CPI_Float32)max;
	col[1] = fvals[1] * (CPI_Float32)max;
	col[2] = fvals[2] * (CPI_Float32)max;
	col[3] = fvals[3] * (CPI_Float32)max;
	tvals[0] *= (CPI_Float32)max;
	tvals[1] *= (CPI_Float32)max;
	tvals[2] *= (CPI_Float32)max;
	tvals[3] *= (CPI_Float32)max;
	delta[0] = ( tvals[0] - col[0] ) / image->myContext.sizeY;
	delta[1] = ( tvals[1] - col[1] ) / image->myContext.sizeY;
	delta[2] = ( tvals[2] - col[2] ) / image->myContext.sizeY;
	delta[3] = ( tvals[3] - col[3] ) / image->myContext.sizeY;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col[0];
			*ptr++ = (T)col[1];
			*ptr++ = (T)col[2];
			*ptr++ = (T)col[3];
		}

		col[0] += delta[0];
		col[1] += delta[1];
		col[2] += delta[2];
		col[3] += delta[3];
	}
}

// handle horizontal alpha-only case

template <class T> static
void make_ramp_x_1( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col, delta;
	T			*ptr;

	// init color stuff
	col = fvals[3] * (CPI_Float32)max;
	tvals[3] *= (CPI_Float32)max;
	delta = ( tvals[3] - col ) / image->myContext.sizeX;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col;
			col += delta;
		}

	}
}

// handle horizontal rgb case

template <class T> static
void make_ramp_x_3( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col[4], delta[4];
	T			*ptr;

	// init color stuff
	col[0] = fvals[0] * (CPI_Float32)max;
	col[1] = fvals[1] * (CPI_Float32)max;
	col[2] = fvals[2] * (CPI_Float32)max;
	tvals[0] *= (CPI_Float32)max;
	tvals[1] *= (CPI_Float32)max;
	tvals[2] *= (CPI_Float32)max;
	delta[0] = ( tvals[0] - col[0] ) / image->myContext.sizeX;
	delta[1] = ( tvals[1] - col[1] ) / image->myContext.sizeX;
	delta[2] = ( tvals[2] - col[2] ) / image->myContext.sizeX;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col[0];
			*ptr++ = (T)col[1];
			*ptr++ = (T)col[2];
			col[0] += delta[0];
			col[1] += delta[1];
			col[2] += delta[2];
		}

	}
}

// handle horizontal rgba output

template <class T> static
void make_ramp_x_4( CPI_Image		*image, 
				    CPI_Float32	 	 fvals[], 
				    CPI_Float32	 	 tvals[],
				    T 			 	 max )
{
	CPI_Int32	 x, y, endX, endY;
	CPI_Float32	 col[4], delta[4];
	T			*ptr;

	// init color stuff
	col[0] = fvals[0] * (CPI_Float32)max;
	col[1] = fvals[1] * (CPI_Float32)max;
	col[2] = fvals[2] * (CPI_Float32)max;
	col[3] = fvals[3] * (CPI_Float32)max;
	tvals[0] *= (CPI_Float32)max;
	tvals[1] *= (CPI_Float32)max;
	tvals[2] *= (CPI_Float32)max;
	tvals[3] *= (CPI_Float32)max;
	delta[0] = ( tvals[0] - col[0] ) / image->myContext.sizeX;
	delta[1] = ( tvals[1] - col[1] ) / image->myContext.sizeX;
	delta[2] = ( tvals[2] - col[2] ) / image->myContext.sizeX;
	delta[3] = ( tvals[3] - col[3] ) / image->myContext.sizeX;

	endX	= image->myContext.offsetX + image->myContext.sizeX;
	endY	= image->myContext.offsetY + image->myContext.sizeY;
	
	// loop through each scanline
	for ( y = image->myContext.offsetY; y < endY; y++ )
	{
		ptr = (T *)cpiGetLine( image, y );
		for ( x = image->myContext.offsetX; x < endX; x++ )
		{
			*ptr++ = (T)col[0];
			*ptr++ = (T)col[1];
			*ptr++ = (T)col[2];
			*ptr++ = (T)col[3];
			col[0] += delta[0];
			col[1] += delta[1];
			col[2] += delta[2];
			col[3] += delta[3];
		}

	}
}



////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
RampCreateOpInstance( void )
{
	rampState *retval = (rampState *)cpiAlloc( sizeof( rampState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where you
// return the memory allocated in the CreateOp routine defined above.


static CPI_Bool
RampDestroyOpInstance( CPI_PrivateData	 handle )
{
	CPI_Bool	 retval = CPI_FALSE;
	rampState	*state	= (rampState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on
//    bit depth

static CPI_Bool
RampProcess( CPI_PrivateData	 handle,
			 CPI_Image			*result,
			 CPI_Image		 	 /*inputs*/[],
			 CPI_Uint32			 /*numInputs*/ )
{
	CPI_Bool	 retval = CPI_FALSE;
	rampState	*state	= (rampState *)handle;

	if ( NULL != state )
	{
		if ( VERTICAL == state->direction )
		{
			if ( 1 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_y_1( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_y_1( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_y_1( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else if ( 3 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_y_3( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_y_3( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_y_3( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else if ( 4 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_y_4( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_y_4( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_y_4( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else
			{
				cpiError( "Invalid number of channels in Ramp" );
			}
		}
		else
		{
			if ( 1 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_x_1( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_x_1( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_x_1( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else if ( 3 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_x_3( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_x_3( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_x_3( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else if ( 4 == result->myContext.channels )
			{
				if ( result->myContext.bitsPerPel == 8 )
					make_ramp_x_4( result, state->fromColor, 
								   state->toColor, MAX8 );
				else if ( result->myContext.bitsPerPel == 16 )
					make_ramp_x_4( result, state->fromColor, 
								   state->toColor, MAX16 );
				else
					make_ramp_x_4( result, state->fromColor, 
								   state->toColor, MAXF );
			}
			else
			{
				cpiError( "Invalid number of channels in Ramp" );
			}
		}

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Ramp parameter(s)" );
	
	return retval;
}


//////////////////////////////////////
// This defines for Rayz the output region of your image
// op. It just passes on the values from the GUI.

static CPI_Bool
RampOutputRegion( CPI_PrivateData		 handle,
				  CPI_ImageContext		*output,
				  CPI_ImageContext		 /*inputs*/[],
				  CPI_Uint32			 /*numInputs*/ )
{	
	CPI_Bool 	 retval	= CPI_FALSE;
	rampState	*state	= (rampState *)handle;

	if ( NULL != state )
	{
		output->sizeX		= state->size[0];
		output->sizeY		= state->size[1];
		output->bitsPerPel	= state->bitdepth;
		output->channels	= state->channels;
		output->offsetX 	= 0;
		output->offsetY 	= 0;
		retval				= CPI_TRUE;
	}
	else
		cpiError( "Couldn't get ramp private data" );
		
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.

static RPI_ImageOpInfo opinfo = 
{
	{
		"ramp",			// op name
		"Ramp Image Operation",
		"Silicon Grail",
		"2.2", NULL, NULL
	},
	RampCreateOpInstance,	// create instance
	RampDestroyOpInstance,	// destroy instance
	RampProcess,			// image execution process
	NULL,					// InputRegion
	RampOutputRegion,		// output region
	NULL,					// control processing
	NULL,					// setup processing
	NULL,					// cleanup processing
	0,						// min inputs
	CPI_FALSE,				// does it multiproc?
	CPI_FALSE				// handle disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "ramp" );
	cpiUnregisterNode( "ramp" );
}
