/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    SimpleParam
//
// COMMENTS: Show how to define and use a single parameter.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>



////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)

typedef struct _simpleState
{
	CPI_Float32	 brightness;
} simpleState;


/////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
SimpleInit( void )
{
	cpiAddFloat( "bright", "Brightness", 1.0, 0.0, 1.0,
				  CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				  "This controls the brightness", 
				  NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
SimpleExec( CPI_PrivateData	 /*handle*/,
			CPI_Float32		 myTime,
			CPI_Uint8		 quality,
			CPI_Uint32		 nodeOutput,
			CPI_Bool	 	 viewerOutput,
			CPI_Float32		 scaleX,
			CPI_Float32   	 scaleY )

{
	CPI_ImageOp		 retval		= NULL;
	CPI_ImageOp 	 myInputs	= NULL;
	
	// get the input
	myInputs = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	// If we have a valid input, then allocate some parameters and
	// pass them to the image op (IM)
	if ( NULL != myInputs )
	{
		// These 4 lines define and allocate a block of data as
		// defined in the earlier typedef
		CPI_PrivateData	 opData		= NULL;
		simpleState		*simpleData	= NULL;

		opData		= cpiCreatePrivateData( "oneparam" );
		simpleData	= (simpleState *)opData;

		if ( NULL != simpleData )
		{
			// get the user parameter value and stuff it into the
			// structure. This structure is then added, along with
			// the input(s) and the image op itself, to the list
			// of operations that RAYZ will execute.
			if ( cpiGetFloat ( &simpleData->brightness, "bright", myTime ) )
			{
				retval = cpiAddImageOp( "oneparam", simpleData, &myInputs, 1 );
			}
			else
			{
				cpiError( "Couldn't get SimpleParam parameter" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
			cpiError( "Couldn't get SimpleParam private data" );
	}
	else
		cpiError( "Missing required input!" );
	
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not defined are set to NULL. Routines
// which are required to be provided are marked as such.

static RPI_NodeInfo ninfo = {
	{							// this structure is required
		"oneparam",				// node name
		"Oneparam",				// menu name
		"Joe Programmer",		// author
		"2.2"					// version
	},
	SimpleInit,			// NodeInit		(required)
	NULL,				// NodeShutDown
	NULL,				// NodeCreateInstance
	NULL,				// NodeDestroyInstance
	NULL,				// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	SimpleExec,			// Node Execute	(required)
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,               // ViewerLabel
	NULL,               // ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.

template <class T> static
void simple( CPI_Image	   &input,
			 CPI_Image	   *result, 
			 CPI_Float32	bright,
			 T 			    max)
{
	CPI_Uint32	 i, x, y, endX, endY;
	CPI_Uint32	 chans = result->myContext.channels;
	T			*pin, *pout;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pin		= (T *)cpiGetLine( &input, y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// multiply each channel by the passed value
			for ( i = 0; i < chans; i++ )
			{
				*pout = (T) (*pin * bright);
				pin++;
				pout++;
			}
		}
	}
}


////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
SimpleCreateOpInstance( void )
{
	simpleState	*retval = (simpleState *)cpiAlloc( sizeof( simpleState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where you
// return the memory allocated in the CreateOp routine defined above.


static CPI_Bool
SimpleDestroyOpInstance( CPI_PrivateData	 handle )
{
	CPI_Bool	 retval = CPI_FALSE;
	simpleState	*state	= (simpleState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth

static CPI_Bool
SimpleProcess( CPI_PrivateData	 handle,
			   CPI_Image		*result,
			   CPI_Image	 	 inputs[],
			   CPI_Uint32	 	 numInputs )
{
	CPI_Bool	 retval	= CPI_FALSE;
	simpleState	*state 	= (simpleState *)handle;

	if ( NULL != state )
	{	
		// pass it to the appropriate routine
		if ( result->myContext.bitsPerPel == 8 )
			simple( inputs[0], result, state->brightness, (CPI_Uint8)255 );
		else if ( result->myContext.bitsPerPel == 16 )
			simple( inputs[0], result, state->brightness, (CPI_Uint16)65535 );
		else
			simple( inputs[0], result, state->brightness, (CPI_Float32)1.0F );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't find SimpleParam private data" );
		
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.
// Required entries are marked, the rest are optional. Fill in
// NULL if there is no entry for an optional function.

static RPI_ImageOpInfo opinfo = 
{
	{						// this structure is required
		"oneparam",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"2.2"				// version
	},
	SimpleCreateOpInstance,	// create op
	SimpleDestroyOpInstance,// destroy op
	SimpleProcess,			// image execution process (required)
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
	NULL,                   // SetupProcessing
	NULL,                   // CleanupProcessing
	1,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE               // Handle disjoint regions?
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "oneparam" );
	cpiUnregisterNode( "oneparam" );
}
