/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    SimpleTwoInput
//
// COMMENTS: Show how to build a node with two inputs. The 2 input
//			 requirement shows up in the RPI_NodeInfo structure,
//			 as well as in the Process() function.
//
*/

#include <stdio.h>
#include <math.h>
#include <CPI.h>


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)

typedef struct _simpleState
{
	CPI_Float32	 dissolve;
} simpleState;


/////////////////////////////////////////
// Init is called when the node is created, and is used
// to define the user input controls

static CPI_Bool
SimpleInit( void )
{
	cpiAddFloat( "dissolve", "Dissolve", 1.0, 0.0, 1.0,
				  CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN,
				  "Amount of dissolve from input A to input B",
				  NULL);

	return CPI_TRUE;
}

//////////////////////////////////////////
// label the node inputs - this is optional
// Input is the number of the input you want to provide a
// label for, starting at 0 - this label shows up in the
// status line when the user mouses over the connector.
// The first letter of each string is used to label the input
// connector.

static const char *
SimpleInputLabel( CPI_PrivateData	 /*handle*/,
				  CPI_Uint32		 input)
{
	static const char	input1[] = {"foreground image"};
	static const char	input2[] = {"background image"};

	if ( input == 0 )
		return input1;
	else
		return input2;
}


//////////////////////////////////////////
// This is called when an image is required
// The basic structure of these routines is:
// 1. get the inputs
// 2. get the parameters from the user interface
// 3. allocate metadata and copy the parameters to it
// 4. add the image op(s) to the list of operations

static CPI_ImageOp
SimpleExec( CPI_PrivateData	 /*handle*/,
			CPI_Float32		 myTime,
			CPI_Uint8		 quality,
			CPI_Uint32		 nodeOutput,
			CPI_Bool	 	 viewerOutput,
			CPI_Float32		 scaleX,
			CPI_Float32   	 scaleY )

{
	CPI_ImageOp		retval	= NULL;
	CPI_ImageOp 	myInputs[2];
	
	// get the two input images
	myInputs[0] = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	myInputs[1] = cpiGetInputOp( 1, myTime, quality, scaleX, scaleY );

	// If we have valid inputs, then allocate some parameters and
	// pass them to the image op (IM)
	if ( myInputs[0] != NULL && myInputs[1] != NULL )
	{
		// These 4 lines define and allocate a block of data as
	    // defined in the earlier typedef
		CPI_PrivateData	 opData		= NULL;
		simpleState		*simpleData	= NULL;

		opData		= cpiCreatePrivateData( "twoinput" );
		simpleData	= (simpleState *)opData;

		if ( NULL != simpleData )
		{
			// get the value from the user and stuff in into the
			// structure. This structure is then added, along with
			// the input(s) and the image op itself, to the list
		    // of operations that RAYZ will execute.
			if ( cpiGetFloat ( &simpleData->dissolve, "dissolve", myTime ) )
			{
				retval = cpiAddImageOp( "twoinput", simpleData, myInputs, 2 );
			}
			else
			{
				cpiError( "Couldn't get SimpleTwoInput parameter" );
				cpiDestroyPrivateData( opData );
			}
		}
		else
			cpiError( "Couldn't get SimpleTwoInput private data" );
	}
	else
		cpiError( "Missing required input(s)" );
	
	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL. Those
// which are required to be provided are marked as such

static RPI_NodeInfo ninfo = {
	{							// this structure is required
		"twoinput",				// node name
		"Twoinput",				// menu name
		"Joe Programmer",		// author
		"2.2"					// version
	},
	SimpleInit,			// NodeInit			(required)
	NULL,				// NodeShutDown
	NULL,				// NodeCreateInstance
	NULL,				// NodeDestroy
	NULL,				// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	SimpleExec,			// Node Execute		(required)
	SimpleInputLabel,	// InputLabel
	NULL,				// OutputLabel
	NULL,               // ViewerLabel
	NULL,               // ViewerOutputs
	2, 					// min inputs
	2,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};



//////////////////////////////////////////////////////////////////////
// Image Op Functions ////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////

// This is the actual image manipulation routine. It is a templated
// function so that we can write it once, and let the compiler build
// a version for each of 8, 16 and float pixel types.

// You are guaranteed that the two images (and the result) will
// all be the same size if the 'disjoint regions' flag is CPI_FALSE
// in the RPI_ImageOpInfo structure below. See later note.

template <class T> static
void simple( CPI_Image	    inputs[],
			 CPI_Image	   *result, 
			 CPI_Float32	dissolve,
			 T 			    max)
{
	CPI_Uint32	 i, x, y, endX, endY;
	CPI_Uint32	 chans = result->myContext.channels;
	T           *pinA, *pinB, *pout;
	CPI_Float32	 oneminus;

	endX	= result->myContext.offsetX + result->myContext.sizeX;
	endY	= result->myContext.offsetY + result->myContext.sizeY;
	oneminus =  1.0 - dissolve;
	
	for ( y = result->myContext.offsetY; y < endY; y++ )
	{
		// Notice that we use cpiGetLine to avoid any sub-image issues
		// as rayz may ask us to process a small part of a larger image
		pinA	= (T *)cpiGetLine( &inputs[0], y );
		pinB	= (T *)cpiGetLine( &inputs[1], y );
		pout	= (T *)cpiGetLine( result, y );

		for ( x = result->myContext.offsetX; x < endX; x++ )
		{
			// do the math
			for ( i = 0; i < chans; i++ )
			{
				*pout = (T) ((*pinA * dissolve) + (*pinB * oneminus));
				pinA++;
				pinB++;
				pout++;
			}
		}
	}
}


////////////////////////////////////////
// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.


static CPI_PrivateData
SimpleCreateOpInstance( void )
{
	simpleState	*retval = (simpleState *)cpiAlloc( sizeof( simpleState ) );

	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where
// you return the memory allocated in the CreateOp routine defined
// above.

static CPI_Bool
SimpleDestroyOpInstance( CPI_PrivateData handle )
{
	CPI_Bool	 retval	= CPI_FALSE;
	simpleState	*state	= (simpleState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


/////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth

static CPI_Bool
SimpleProcess( CPI_PrivateData	 handle,
			   CPI_Image		*result,
			   CPI_Image	 	 inputs[],
			   CPI_Uint32	 	 numInputs )
{
	CPI_Bool	 retval = CPI_FALSE;
	simpleState	*state	= (simpleState *)handle;

	if ( NULL != state )
	{
		// call the routine with parameters
		if ( result->myContext.bitsPerPel == 8 )
			simple( inputs, result, state->dissolve, (CPI_Uint8)255 );
		else if ( result->myContext.bitsPerPel == 16 )
			simple( inputs, result, state->dissolve, (CPI_Uint16)65535 );
		else
			simple( inputs, result, state->dissolve, (CPI_Float32)1.0F );

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Simple private data" ); 
		
	return retval;
}


/////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.
// Required entries are marked, the rest are optional. Fill in
// NULL if there is no entry for an optional function.

static RPI_ImageOpInfo opinfo = 
{
	{						// this structure is required
		"twoinput",			// command name
		"",			        // not used
		"Joe Programmer",	// author
		"2.2"				// version
	},
	SimpleCreateOpInstance,	// CreateOpInstance
	SimpleDestroyOpInstance,// DestroyOpInstance
	SimpleProcess,			// image execution process	 (required)
	NULL,					// InputRegion
	NULL,					// OutputRegion,
	NULL,                   // ControlProcessing
    NULL,                   // SetupProcessing
    NULL,                   // CleanupProcessing
	2,						// min number of inputs required
	CPI_TRUE,				// Can this plugin multiproc?
	CPI_FALSE               // Handle disjoint regions?
	// the Disjoint regions flag means this:
	// if == CPI_FALSE, Rayz will compute the common pixels 
	// (intersection) of all your inputs, and just pass those
	// as inputs to the Process() routine.
	// The area outside the intersection will be set to the
	// Backing Color defined in the background's image context. 
	// If you want something other than that to happen, you 
	// need to define the ControlProcessing function. See the 
	// example called NewMult.C
	// if == CPI_TRUE, you will get handed the complete inputs,
	// and you are on your own as to how to process them
};


////////////////////////////////////
// register the node and the image op
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval =  cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "twoinput" );
	cpiUnregisterNode( "twoinput" );
}
