/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  710 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:	Stats
//
// COMMENTS: dump statistics about an image to the node panel. This 
// 			 demonstrates interacting with an image
//
*/

#include <stdio.h>
#include <math.h>

#include <CPI.h>


////////////////////////////////////////
// In order to pass parameters between the COP level and
// the IM level, we define a structure of values. An
// instance of this structure will be passed down to the
// IM (IMage operation)


typedef struct _statState
{
	CPI_Float32	 min[4];
	CPI_Float32	 max[4];
	CPI_Float32	 avg[4];
	CPI_Float32	 avgDeviation[4];
	CPI_Float32	 stdDeviation[4];
	CPI_Float32	 variance[4];
} statState;


////////////////////////////////////////


static CPI_Float32	mindefs[] = { 0.0F, 0.0F, 0.0F, 0.0F };
static CPI_Float32	maxdefs[] = { 1.0F, 1.0F, 1.0F, 1.0F };
static CPI_Float32	avgdefs[] = { 0.5F, 0.5F, 0.5F, 0.5F };


///////////////////////////////////////////////////////////////////////////////
// Node Functions /////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
StatInit( void )
{
	cpiAddColorGroup( "minimum",
					  "Minimum",
					  mindefs,
					  4,
					  CPI_PARM_DEFAULT,
					  "Minimum color of the image",
					  NULL );
	cpiAddColorGroup( "maximum",
					  "Maximum",
					  maxdefs,
					  4,
					  CPI_PARM_DEFAULT,
					  "Maximum color of the image",
					  NULL );
	cpiAddColorGroup( "average",
					  "Average",
					  avgdefs,
					  4,
					  CPI_PARM_DEFAULT,
					  "Average color of the image",
					  NULL );

	return CPI_TRUE;
}


////////////////////////////////////////


static CPI_Bool
StatShutDown( void )
{
	return CPI_TRUE;
}


////////////////////////////////////////


static CPI_ImageOp
StatExec( CPI_PrivateData	 /*handle*/,
		  CPI_Float32		 myTime,
		  CPI_Uint8			 quality,
		  CPI_Uint32		 /*output*/,
		  CPI_Bool			 /*viewerOutput*/,
		  CPI_Float32		 scaleX,
		  CPI_Float32		 scaleY )
{
	CPI_ImageOp retval = NULL;
	
	// in this case, the exec simply returns the input image
	retval = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );
	
	return retval;
}


////////////////////////////////////////


static RPI_NodeInfo ninfo = {
	{
		"stats",
		"Stats",
		"Silicon Grail",
		"2.2"
	},
	StatInit,
	StatShutDown,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	StatExec,
	NULL,
	NULL,
	NULL,
	NULL,
	1,
	1,
	1,
	1,
	CPI_FALSE
};

//////////////////////////////////////////////////
// Image Op Functions ////////////////////////////
//////////////////////////////////////////////////


// This is called when a new node of this type is created.
// Typically, this is the place that allocates a particular
// instance of the parameter-holding structure defined earlier.

static CPI_PrivateData
StatsCreateOpInstance( void )
{
	statState *retval = (statState *)cpiAlloc( sizeof( statState ) );

	for ( CPI_Int32 i = 0; i < 4; ++i )
	{
		retval->min[i]	= 1.0F;
		retval->max[i]	= 0.0F;
		retval->avg[i]	= 0.0F;
	}
	
	return (CPI_PrivateData)retval;
}


////////////////////////////////////////
// This is called when the node is deleted. Typically, this is where you
// return the memory allocated in the CreateOp routine defined above.

static CPI_Bool
StatsDestroyOpInstance( CPI_PrivateData handle )
{
	CPI_Bool	 retval	= CPI_FALSE;
	statState	*state	= (statState *)handle;

	if ( NULL != state )
	{
		cpiFree( state );
		retval = CPI_TRUE;
	}

	return retval;
}


////////////////////////////////////////


#define MAX8	(CPI_Uint8)255
#define MAX16	(CPI_Uint16)65535
#define MAXF	(CPI_Float32)1.0F


////////////////////////////////////////


template <class T> static
void calcStats( CPI_Image	 &input,
				CPI_Image	 *result,
				statState	 *state,
				T			  max )
{
	CPI_Int32	 x, y, c, endX, endY;
	CPI_Float32	 sum[4];
	T			*dataPtr	= NULL;

	sum[0]	= 0.0F;
	sum[1]	= 0.0F;
	sum[2]	= 0.0F;
	sum[3]	= 0.0F;
	
	endX	= input.myContext.offsetX + input.myContext.sizeX;
	endY	= input.myContext.offsetY + input.myContext.sizeY;

	for ( y = input.myContext.offsetY; y < endY; ++y )
	{
		dataPtr = (T*)cpiGetLine( &input, y );

		for ( x = 0; x < input.myContext.sizeX; ++x )
		{
			for ( c = 0; c < input.myContext.channels; ++c )
			{
				CPI_Float32	 fltData	= (CPI_Float32)*dataPtr++ / max;

				if ( fltData < state->min[c] )
					state->min[c] = fltData;
				if ( fltData > state->max[c] )
					state->max[c] = fltData;

				sum[c] += fltData;
			}
		}
	}

	for ( c = 0; c < input.myContext.channels; ++c )
	{
		state->avg[c] = sum[c] / ( input.myContext.sizeX *
								   input.myContext.sizeY );
	}
}


////////////////////////////////////////
// This is called to execute the image op
// The basic structure of these routines is
// 1. unload parameter values from metadata
// 2. call the appropriate routine depending on bit depth


static CPI_Bool
StatsProcess( CPI_PrivateData	 handle,
			  CPI_Image			*result,
			  CPI_Image			 inputs[],
			  CPI_Uint32		 numInputs )
{
	CPI_Bool	 retval	= CPI_FALSE;
	statState	*state	= (statState *)handle;
	
	if ( NULL != state )
	{
		if ( result->myContext.bitsPerPel == 8 )
			calcStats( inputs[0], result, state, MAX8 );
		else if ( result->myContext.bitsPerPel == 16 )
			calcStats( inputs[0], result, state, MAX16 );
		else
			calcStats( inputs[0], result, state, MAXF );					   

		retval = CPI_TRUE;
	}
	else
		cpiError( "Couldn't get Stats private data" );

	return retval;
}


////////////////////////////////////////
// This structure defines the image op (IM) that you
// make. It is required if you are creating your own IM.
// Required entries are marked, the rest are optional. Fill in
// NULL if there is no entry for an optional function.

static RPI_ImageOpInfo opinfo =
{
	{						// this is required
		"stats",
		"",
		"silicon grail",
		"2.2"
	},
	StatsCreateOpInstance,
	StatsDestroyOpInstance,
	StatsProcess,			// required
	NULL,					// Input Region
	NULL,					// Output Region
	NULL,					// Control Processing
	NULL,					// set up processing
	NULL,					// cleanup processing
	1,						// min number of inputs required
	CPI_FALSE,				// can it multiproc?
	CPI_FALSE,				// handle disjoint regions?
};

///////////////////////////////////////////////////////////
// Overlay Functions //////////////////////////////////////
///////////////////////////////////////////////////////////


static void
setStats( CPI_Image			*result,
		  CPI_PrivateData	 callerData,
		  CPI_PrivateData	 dataList[],
		  CPI_Uint32		 numOfData,
		  CPI_Float32		 time )
{
	if ( numOfData == 1 )
	{
		statState	*state	= (statState *)dataList[0];
		
		if ( NULL != state )
		{
			cpiSetFloat( state->min[0], "minimum.red", time, CPI_FALSE );
			cpiSetFloat( state->min[1], "minimum.green", time, CPI_FALSE );
			cpiSetFloat( state->min[2], "minimum.blue", time, CPI_FALSE );
			cpiSetFloat( state->min[3], "minimum.alpha", time, CPI_FALSE );
			cpiSetFloat( state->max[0], "maximum.red", time, CPI_FALSE );
			cpiSetFloat( state->max[1], "maximum.green", time, CPI_FALSE );
			cpiSetFloat( state->max[2], "maximum.blue", time, CPI_FALSE );
			cpiSetFloat( state->max[3], "maximum.alpha", time, CPI_FALSE );
			cpiSetFloat( state->avg[0], "average.red", time, CPI_FALSE );
			cpiSetFloat( state->avg[1], "average.green", time, CPI_FALSE );
			cpiSetFloat( state->avg[2], "average.blue", time, CPI_FALSE );
			cpiSetFloat( state->avg[3], "average.alpha", time, CPI_FALSE );
		}
		else
			printf( "Incorrect private data!" );
	}	
	else
		printf( "I only expected to get one piece of private data!" );
}


////////////////////////////////////////


static void
computeStats( const char		*nodename,
			  const char		*parmname,
			  CPI_Float32		 time,
			  CPI_PrivateData	 data )
{
	CPI_ImageOp	lastOp	= NULL;

	lastOp = cpiCreateNewExecList();

	if ( NULL != lastOp )
	{
		CPI_PrivateData	 opData = cpiCreatePrivateData( "stats" );
		
		lastOp = cpiAddImageOp( "stats", opData, &lastOp, 1 );

		cpiStartExecution( setStats, NULL, lastOp );
	}
	else
		printf("lastOp is NULL!\n" );
}


////////////////////////////////////////
// This is where we do any initialization, including defining
// a toolbar


static CPI_Bool
StatInitOverlay( void )
{
	cpiStartToolbar();
	cpiAddButton( "compute",
				  "Compute Values",
				  computeStats,
				  CPI_PARM_NO_LABEL,
				  NULL,
				  NULL );
	cpiEndToolbar( "statbar", "Stats", CPI_WINDOW_BOTTOM, true, false );

	return CPI_TRUE;	
}


////////////////////////////////////////


static void
StatRender( CPI_PrivateData handle, CPI_Bool isPicking )
{
	// We don't want to draw anything
}


////////////////////////////////////////


static RPI_OverlayInfo oinfo = {
	{
		"stats",
		"Stats",
		"Silicon Grail",
		"v1.0"
	},
	StatInitOverlay,		//Init
	NULL,					//Shutdown
	NULL,					//CreateInstance
	NULL,					//DestroyInstance
	NULL,					//SourceChanged
	NULL,					//FullSizeChanged
	NULL,					//ParmChanged
	NULL,					//SoureOutOfDate
	StatRender,				//Render
	NULL,					//LeftMouseDown
	NULL,					//LeftMouseDrag
	NULL,					//LeftMouseUp
	NULL,					//MouseEnter
	NULL,					//MouseDrag
	NULL,					//MouseExit
	"stats",
	CPI_FALSE
};


////////////////////////////////////////


CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool	 retval = CPI_TRUE;

	retval = cpiRegisterImageOp( &opinfo );

	if ( retval )
		retval = cpiRegisterNode( &ninfo );

	if ( retval )
		retval = cpiRegisterOverlay( &oinfo );

	return retval;
}


////////////////////////////////////////


CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterImageOp( "stats" );
	cpiUnregisterNode( "stats" );
	cpiUnregisterOverlay( "stats" );
}
