/*
// PROPRIETARY INFORMATION.  This software is proprietary to
// Silicon Grail Corporation, and is not to be reproduced,
// transmitted, or disclosed in any way without written permission.
//
// Produced by:
//  Silicon Grail Corporation
//  736 N. Seward
//  Hollywood, CA
//  USA 90038
//  323-871-9100
//
// NAME:    Widgets
//
// COMMENTS: Demonstrate the widgets available to the plugin programmer
//
*/

#include <string.h>
#include <stdio.h>
#include <math.h>
#include <CPI.h>
#include "Widgets.h"

/////////////////////////////////////
// masks for channel selection

#define RED_MASK	0x01
#define GREEN_MASK	0x02
#define BLUE_MASK	0x04
#define ALPHA_MASK	0x08

// string size constant
#define MAXSIZE		256
static char			showstring[MAXSIZE];


/////////////////////////////////////
// Callback for button push in node panel
// This is reached when the user pushes the "push me" button
// defined in the init routine.
// In this situation, the value is not set by the user directly, so
// the CPI_FALSE flag prevents that value from being saved into the
// resulting .rayz file. 

static void
WidgetCallback( const char 			*nodename,
			    const char 			*parmname,
				CPI_Float32 	 	 time,
				CPI_PrivateData		 data )
{
	sprintf (showstring, 
	"Button callback for node '%s'\nparameter '%s'\nat time %5.5f\n",
			nodename, parmname, time );
	cpiSetString( showstring, "text_entry", time, CPI_FALSE );
	cpiSetInteger( (CPI_Int32)time, "integer", time, CPI_FALSE );
}


////////////////////////////////////
// Callback for button pushed in Var List
// This is reached when the user wants to add another parameter
// to the variable length list of params. An example of this
// is track, where you add track points
// An element can contain any number of any kind of parameters

static void
addNewElement( void )
{
	cpiAddInteger( "new_thing", "new thing", 0, 0, 10, 
					CPI_PARM_DEFAULT,
				   "This was added to the var list", NULL );
}


/////////////////////////////////////////
// define a menu to use
static const char *mymenu[]=
{
	"lion",
	"tiger",
	"bear",
	NULL
};


////////////////////////////////////////////
// define an array of icon names
static const char *icons[] =
{
	"Add",
	"Subtract",
	"Multiply",
	"Divide",
	NULL
};


////////////////////////////////////////////
// The init routine is called once, and typically is used to
// define the user input controls

static CPI_Bool
WidgetInit( void )
{
	// add a menu
	cpiAddMenu( "menu", "Menu", 0, mymenu, 
			    "Choose which type of animal from menu", NULL );

	// add an integer value
	// (name, default_value, min_value, max_value, flags, help, html)
	cpiAddInteger( "integer", "Integer", 1, 0, 100,
				    CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN | CPI_PARM_CLAMPMAX, 
				   "Integer parameter", NULL );

	// add a pair of integer parameters
	cpiAddIntPair( "int_pair", "Int pair", 
				   "w", "Width", 720, "h", "Height", 486, 
					1, 4096, CPI_PARM_CLAMPMIN, 
				   "Pair of integer parameters", NULL );

	// add a float value
	cpiAddFloat( "float", "Float", 1.0, 0.0, 1.0,
				  CPI_PARM_ANIMATABLE | CPI_PARM_CLAMPMIN, 
				 "Floating point parameter", NULL );

	// add a pair of float values
	cpiAddFloatPair( "float_pair", "Float pair", 
					 "x", "X", 1.0, "y", "Y", 0.75, 
					  0.0, 1.0, CPI_PARM_DEFAULT, 
					 "Pair of floating point values", NULL );

	// add a file name with browser
	cpiAddFile( "file", "File", "/tmp/myfile", 
				"This is a file name", NULL );
					
	// add a single input string
	cpiAddString( "string", "String", "default", CPI_FALSE, 
				   CPI_PARM_DEFAULT, 
				  "This is an input string", NULL );

	// add a text entry block
	cpiAddString( "text_entry", "Text Entry", "Push Callback Button", 
				   CPI_TRUE, CPI_PARM_DEFAULT, 
				  "This is a block of text to input", NULL );

	// add a read-only piece of text
	cpiAddDisplayString( "display", "Display text", "rest of string");

	// add a toggle button
	// (name, default_value, flags, help, html)
	cpiAddToggle( "toggle", "Toggle", CPI_TRUE, CPI_PARM_DEFAULT, 
			      "Toggle button which defaults to true", NULL );

	// add a toggle with an icon
	// We define two icons, off and on, and attach one to the
	// "normal" state and the other to the "selected" state
	// To be really slick, we could/should have separate icon
	// versions for located, disabled, etc
	cpiRegisterIconData( "Off", 24, 18, off_data, 255, 0, 255 );
	cpiRegisterIconData( "On", 24, 18, on_data, 255, 0, 255 );
	cpiDefineIcon( "Off", "Off",
				   NULL, NULL,  
				   "On", NULL, NULL, 255, 0, 255 );
	cpiAddIconToggle( "itoggle", "Off", CPI_FALSE, CPI_PARM_DEFAULT,
					  "Toggle button with icon", NULL );

	// add a button which does something immediately when pressed
	cpiAddButton( "push_me", "Callback Button", WidgetCallback, 
				   CPI_PARM_DEFAULT,
				  "This is a callback button", NULL );

	// add a button with an icon
	cpiRegisterIconData( "Atomic", 32, 24, atomic_data, 255, 0, 255 );
	cpiDefineIcon( "Atomic", "Atomic",
				   NULL, NULL, NULL, NULL, NULL, 255, 0, 255 );
	cpiAddIconButton( "ibutton", "Icon Button", "Atomic", 
					  WidgetCallback, CPI_PARM_DEFAULT,
					  "Callback button with icon", NULL );

	// add a button strip
	cpiAddButtonStrip( "button_strip", "Button Strip", 
	                    0, CPI_PARM_DEFAULT, mymenu, "", NULL );

	// add an icon button strip
	// First, register the icon data contained in the header file
	// Icons.h
	cpiRegisterIconData( icons[0], 24, 24, add_data,      255, 0, 255 );
	cpiRegisterIconData( icons[1], 24, 24, subtract_data, 255, 0, 255 );
	cpiRegisterIconData( icons[2], 24, 24, multiply_data, 255, 0, 255 );
	cpiRegisterIconData( icons[3], 24, 24, divide_data,   255, 0, 255 );
	// Then, define the icons by name
	for ( CPI_Int32 i = 0; i < 4; ++i )
		cpiDefineIcon( icons[i], icons[i],
					   NULL, NULL,  NULL, NULL, NULL, 255, 0, 255 );
	// finally, define the strip
	cpiAddIconButtonStrip( "icon_strip", "Icon Button Strip", 0,
							CPI_PARM_DEFAULT, icons, 
							"This is a button strip with icons", NULL );

	// add a nested group of parameters
	cpiStartGroup( "group1", "Group 1", CPI_PARM_DEFAULT,
				   "This is a nested group of parameters", NULL );
	cpiAddFloat( "float2", "Float2", 0.0, -1.0, 1.0,
				  CPI_PARM_ANIMATABLE,
				 "Second floating point parameter", NULL );
	cpiEndGroup();

	// add a color group
	CPI_Float32		def[4] = {0.5, 0.2, 0.9, 1.0};
	cpiAddColorGroup( "color", "Color", def, 4, CPI_PARM_DEFAULT,
					  "This is a color selector", NULL );

	// add a hue adjustment circle
	cpiAddHueCircle( "hue", "Hue Circle", 0.0, CPI_PARM_DEFAULT,
					 "A hue adjustment circle", NULL );

	// add a channel group (integer)
	// groups may not have more than 5 parameters
	cpiAddIntRGBAOGroup( "int_group", "Int Group", 
						  0, 0, 100, 3, CPI_PARM_DEFAULT,
						 "A related group of integer channels", NULL );

	// add a channel group (floating point)
	// groups may not have more than 5 parameters
	cpiAddFloatRGBAOGroup( "float_group", "Float Group", 
							0.0, -1.0, 1.0, 5, CPI_PARM_DEFAULT,
						   "A related group of float channels", NULL );

	// create a variable-length list of elements
	cpiStartVarList();
	cpiDefineVarListGroup( "add_new", "Add New", (void *)addNewElement );
	cpiEndVarList( "var_list", "Var List", CPI_PARM_DEFAULT, CPI_TRUE );

	// add a set of channel selector buttons
	cpiAddChannelSelect( "Select the channels to operate on", NULL );

	// add a frame range indicator
	// this only has meaning for source nodes (nodes with no inputs)
	cpiAddSequenceRange( 1, 100, CPI_PARM_DEFAULT,
						 "Range of frames to create", NULL );

	return CPI_TRUE;
}


//////////////////////////////////////////
// This is called when the node is actually executed
// This demonstrates getting the widget values from the user, and then
// passing them into the image op via metadata

static CPI_ImageOp
WidgetExec( CPI_PrivateData	 /*handle*/,
			CPI_Float32		 myTime,
			CPI_Uint8		 quality,
			CPI_Uint32		 nodeOutput,
			CPI_Bool	 	 viewerOutput,
			CPI_Float32		 scaleX,
			CPI_Float32   	 scaleY )

{
	CPI_Float32		floatval, xval, yval, red, green, blue, alpha;
	CPI_Float32		hue, overall, other;
	CPI_Int32		chan, intval, menuval, width, height, toggle;
	CPI_ImageOp		retval	= NULL;
	char			newstr[MAXSIZE], filename[MAXSIZE];
	
	printf("-----Show Selected Parameter Values----\n");

	// get the current choice from a menu
	// result is numbered from 0, so 0 == first element, 1 == second
	// in this example, 0 == lion, 1 == tiger, 2 == bear
	cpiGetInteger ( &menuval, "menu", myTime );

	// get the value from an integer parameter
	cpiGetInteger( &intval, "integer", myTime );

	// get the value from a float parameter
	cpiGetFloat ( &floatval, "float", myTime );

	// get the values from an integer pair
	// create the parameter name by joining the main name with
	// the specific name, separated by a '.'
	cpiGetInteger( &width, "int_pair.w", myTime );
	cpiGetInteger( &height, "int_pair.h", myTime );

	// get the values from a float pair
	cpiGetFloat( &xval, "float_pair.x", myTime );
	cpiGetFloat( &yval, "float_pair.y", myTime );

	// get the string from a single string parameter
	// the same call is used for a block of text
	cpiGetString( newstr, MAXSIZE, "string", myTime );
	printf("string:\t\t\t%s\n", newstr);

	// the same call is used for a file name field, but note
	// that it has an additional subfield to identify it
	cpiGetString( filename, MAXSIZE, "file.file_id", myTime );
	printf("fileid:\t\t\t%s\n", filename);

	// check the setting of a toggle button
	// 0 == off, 1 == on
	cpiGetInteger( &toggle, "toggle", myTime );
	printf("toggle:\t\t\t%s\n", (toggle ? "set" : "unset"));

	// check the setting of the icon toggle - same as for the toggle
	// 0 == off, 1 == on
	cpiGetInteger( &toggle, "itoggle", myTime );
	printf("icon toggle:\t\t%s\n", (toggle ? "set" : "unset"));

	// get a parameter value from inside a group folder
	// must specify the complete path, ie group.param
	cpiGetFloat( &floatval, "group1.float2", myTime);

	// find out which button in the strip is pushed
	cpiGetInteger( &menuval, "button_strip", myTime );
	printf("menu button:\t\t%d\n", menuval );

	// find out which icon button in the strip is pushed
	cpiGetInteger( &menuval, "icon_strip", myTime );
	printf("icon button:\t\t%d\n", menuval );

	// get the color values back from a color selector
	cpiGetFloat( &red, "color.red", myTime );
	cpiGetFloat( &green, "color.green", myTime );
	cpiGetFloat( &blue, "color.blue", myTime );
	cpiGetFloat( &alpha, "color.alpha", myTime );

	// get the value from a hue circle
	cpiGetFloat( &hue, "hue", myTime );

	// get the values from a float group
	// the integer group is analogous
	cpiGetFloat( &overall, "float_group.float_group", myTime );
	cpiGetFloat( &red, "float_group.red", myTime );
	cpiGetFloat( &green, "float_group.green", myTime );
	cpiGetFloat( &blue, "float_group.blue", myTime );
	cpiGetFloat( &alpha, "float_group.alpha", myTime );
	cpiGetFloat( &other, "float_group.other", myTime );

	// get a value from a Var List element
	// each element is addressed with (n) where n is the position
	// in the list created by the user. n begins at 0
	cpiGetInteger( &intval, "var_List(0).add_new.new_thing", myTime);

	// See which channels are selected
	cpiGetChannelSelect( &chan );
	printf("Selected channels:\t");
	if ( chan & RED_MASK )
		printf("Red ");
	if ( chan & GREEN_MASK )
		printf("Green ");
	if ( chan & BLUE_MASK )
		printf("Blue ");
	if ( chan & ALPHA_MASK )
		printf("Alpha ");
	printf("\n");

	// just return the input image
	retval = cpiGetInputOp( 0, myTime, quality, scaleX, scaleY );

	return retval;
}


////////////////////////////////////////////
// This structure, which is required, defines to Rayz
// the various routines which are part of this plugin.
// Routines which are not required are set to NULL

static RPI_NodeInfo ninfo = {
	{
		"widgets",				// node name
		"Widgets",				// menu name
		"Joe Programmer",		// author
		"2.2"					// version
	},
	WidgetInit,			// NodeInit
	NULL,				// NodeShutDown
	NULL,				// NodeCreateInstance
	NULL,				// NodeDestroyInstance
	NULL,				// ParmChanged
	NULL,				// InputAdded
	NULL,				// InputChanged
	NULL,				// InputRemoved
	NULL,				// GetRegion
    NULL,				// GetFullSize
	NULL,				// GetRange
	NULL,				// GetErrors
	WidgetExec,			// Node Execute
	NULL,				// InputLabel
	NULL,				// OutputLabel
	NULL,				// ViewerLabel
	NULL,				// ViewerOutputs
	1, 					// min inputs
	1,					// max inputs
	1,					// min outputs
	1,					// max outputs
	CPI_FALSE			// is growable (eg Mcomp)
};

////////////////////////////////////
// register the node (note, not the image op, because there isn't one)
// This routine and the next are required, and they should
// be declared as shown, in order to work in both NT and
// linux/IRIX

CPIDSOEXPORT CPI_Bool
rpiPluginInit( void )
{
	CPI_Bool retval = CPI_TRUE;

	retval = cpiRegisterNode( &ninfo );

	return retval;	
}


//////////////////////////////////////////////
// unregister node and image op

CPIDSOEXPORT void
upiPluginCleanup( void )
{
	cpiUnregisterNode( "widgets" );
}
